import unittest
import numpy as np
from .. import Mesh, Maps, Utils, SolverLU
from .. import EM
import sys
from scipy.constants import mu_0

FLR = 1e-20  # "zero", so if residual below this --> pass regardless of order
CONDUCTIVITY = 1e1
MU = mu_0
freq = 5e-1


def getFDEMProblem(fdemType, comp, SrcList, freq, useMu=False, verbose=False):
    cs = 10.0
    ncx, ncy, ncz = 0, 0, 0
    npad = 8
    hx = [(cs, npad, -1.3), (cs, ncx), (cs, npad, 1.3)]
    hy = [(cs, npad, -1.3), (cs, ncy), (cs, npad, 1.3)]
    hz = [(cs, npad, -1.3), (cs, ncz), (cs, npad, 1.3)]
    mesh = Mesh.TensorMesh([hx, hy, hz], ["C", "C", "C"])

    if useMu is True:
        mapping = [("sigma", Maps.ExpMap(mesh)), ("mu", Maps.IdentityMap(mesh))]
    else:
        mapping = Maps.ExpMap(mesh)

    x = (
        np.array(
            [np.linspace(-5.0 * cs, -2.0 * cs, 3), np.linspace(5.0 * cs, 2.0 * cs, 3)]
        )
        + cs / 4.0
    )  # don't sample right by the source, slightly off alignment from either staggered grid
    XYZ = Utils.ndgrid(x, x, np.linspace(-2.0 * cs, 2.0 * cs, 5))
    Rx0 = getattr(EM.FDEM.Rx, "Point_" + comp[0])
    if comp[2] == "r":
        real_or_imag = "real"
    elif comp[2] == "i":
        real_or_imag = "imag"
    rx0 = Rx0(XYZ, comp[1], "imag")

    Src = []

    for SrcType in SrcList:
        if SrcType == "MagDipole":
            Src.append(
                EM.FDEM.Src.MagDipole([rx0], freq=freq, loc=np.r_[0.0, 0.0, 0.0])
            )
        elif SrcType == "MagDipole_Bfield":
            Src.append(
                EM.FDEM.Src.MagDipole_Bfield([rx0], freq=freq, loc=np.r_[0.0, 0.0, 0.0])
            )
        elif SrcType == "CircularLoop":
            Src.append(
                EM.FDEM.Src.CircularLoop([rx0], freq=freq, loc=np.r_[0.0, 0.0, 0.0])
            )
        elif SrcType == "RawVec":
            if fdemType == "e" or fdemType == "b":
                S_m = np.zeros(mesh.nF)
                S_e = np.zeros(mesh.nE)
                S_m[
                    Utils.closestPoints(mesh, [0.0, 0.0, 0.0], "Fz")
                    + np.sum(mesh.vnF[:1])
                ] = 1e-3
                S_e[
                    Utils.closestPoints(mesh, [0.0, 0.0, 0.0], "Ez")
                    + np.sum(mesh.vnE[:1])
                ] = 1e-3
                Src.append(
                    EM.FDEM.Src.RawVec(
                        [rx0], freq, S_m, mesh.getEdgeInnerProduct() * S_e
                    )
                )

            elif fdemType == "h" or fdemType == "j":
                S_m = np.zeros(mesh.nE)
                S_e = np.zeros(mesh.nF)
                S_m[
                    Utils.closestPoints(mesh, [0.0, 0.0, 0.0], "Ez")
                    + np.sum(mesh.vnE[:1])
                ] = 1e-3
                S_e[
                    Utils.closestPoints(mesh, [0.0, 0.0, 0.0], "Fz")
                    + np.sum(mesh.vnF[:1])
                ] = 1e-3
                Src.append(
                    EM.FDEM.Src.RawVec(
                        [rx0], freq, mesh.getEdgeInnerProduct() * S_m, S_e
                    )
                )

    if verbose:
        print(f"  Fetching {fdemType!s} problem")

    if fdemType == "e":
        survey = EM.FDEM.Survey(Src)
        prb = EM.FDEM.Problem3D_e(mesh, sigmaMap=mapping)

    elif fdemType == "b":
        survey = EM.FDEM.Survey(Src)
        prb = EM.FDEM.Problem3D_b(mesh, sigmaMap=mapping)

    elif fdemType == "j":
        survey = EM.FDEM.Survey(Src)
        prb = EM.FDEM.Problem3D_j(mesh, sigmaMap=mapping)

    elif fdemType == "h":
        survey = EM.FDEM.Survey(Src)
        prb = EM.FDEM.Problem3D_h(mesh, sigmaMap=mapping)

    else:
        raise NotImplementedError()
    prb.pair(survey)

    try:
        from pymatsolver import Pardiso

        prb.Solver = Pardiso
    except ImportError:
        prb.Solver = SolverLU
    # prb.solverOpts = dict(check_accuracy=True)

    return prb


def crossCheckTest(
    SrcList,
    fdemType1,
    fdemType2,
    comp,
    addrandoms=False,
    useMu=False,
    TOL=1e-5,
    verbose=False,
):

    l2norm = lambda r: np.sqrt(r.dot(r))

    prb1 = getFDEMProblem(fdemType1, comp, SrcList, freq, useMu, verbose)
    mesh = prb1.mesh
    print(
        "Cross Checking Forward: {!s}, {!s} formulations - {!s}".format(
            fdemType1, fdemType2, comp
        )
    )

    logsig = np.log(np.ones(mesh.nC) * CONDUCTIVITY)
    mu = np.ones(mesh.nC) * MU

    if addrandoms is True:
        logsig += np.random.randn(mesh.nC) * np.log(CONDUCTIVITY) * 1e-1
        mu += np.random.randn(mesh.nC) * MU * 1e-1

    if useMu is True:
        m = np.r_[logsig, mu]
    else:
        m = logsig

    survey1 = prb1.survey
    d1 = survey1.dpred(m)

    if verbose:
        print("  Problem 1 solved")

    prb2 = getFDEMProblem(fdemType2, comp, SrcList, freq, useMu, verbose)

    survey2 = prb2.survey
    d2 = survey2.dpred(m)

    if verbose:
        print("  Problem 2 solved")

    r = d2 - d1
    l2r = l2norm(r)

    tol = np.max([TOL * (10 ** int(np.log10(0.5 * (l2norm(d1) + l2norm(d2))))), FLR])
    print(l2norm(d1), l2norm(d2), l2r, tol, l2r < tol)
    return l2r < tol



import SimPEG
from SimPEG.EM.Utils import omega
from SimPEG.EM.Base import BaseEMSurvey
from scipy.constants import mu_0
from SimPEG.Utils import Zero, Identity
from . import SrcFDEM as Src
from . import RxFDEM as Rx


class Survey(BaseEMSurvey):
    """
    Frequency domain electromagnetic survey

    :param list srcList: list of FDEM sources used in the survey
    """

    srcPair = Src.BaseFDEMSrc
    rxPair = Rx.BaseRx

    def __init__(self, srcList, **kwargs):
        # Sort these by frequency
        self.srcList = srcList
        BaseEMSurvey.__init__(self, srcList, **kwargs)

        _freqDict = {}
        for src in srcList:
            if src.freq not in _freqDict:
                _freqDict[src.freq] = []
            _freqDict[src.freq] += [src]

        self._freqDict = _freqDict
        self._freqs = sorted([f for f in self._freqDict])

    @property
    def freqs(self):
        """Frequencies"""
        return self._freqs

    @property
    def nFreq(self):
        """Number of frequencies"""
        return len(self._freqDict)

    @property
    def nSrcByFreq(self):
        """Number of sources at each frequency"""
        if getattr(self, "_nSrcByFreq", None) is None:
            self._nSrcByFreq = {}
            for freq in self.freqs:
                self._nSrcByFreq[freq] = len(self.getSrcByFreq(freq))
        return self._nSrcByFreq

    def getSrcByFreq(self, freq):
        """
        Returns the sources associated with a specific frequency.
        :param float freq: frequency for which we look up sources
        :rtype: dictionary
        :return: sources at the sepcified frequency
        """
        assert freq in self._freqDict, "The requested frequency is not in this survey."
        return self._freqDict[freq]



from .. import Survey, Maps


class LinearSurvey(Survey.BaseSurvey):
    """Base Magnetics Survey"""

    rxLoc = None  #: receiver locations
    rxType = None  #: receiver type
    components = ["gz"]

    def __init__(self, srcField, **kwargs):
        self.srcField = srcField
        Survey.BaseSurvey.__init__(self, **kwargs)

    def eval(self, u):
        return u

    @property
    def nD(self):
        return len(self.rxLoc) * len(self.components)

    @property
    def rxLoc(self):
        return self.srcField.rxList[0].locs

    @property
    def nRx(self):
        return self.srcField.rxList[0].locs.shape[0]

    @property
    def Qfx(self):
        if getattr(self, "_Qfx", None) is None:
            self._Qfx = self.prob.mesh.getInterpolationMat(self.rxLoc, "Fx")
        return self._Qfx

    @property
    def Qfy(self):
        if getattr(self, "_Qfy", None) is None:
            self._Qfy = self.prob.mesh.getInterpolationMat(self.rxLoc, "Fy")
        return self._Qfy

    @property
    def Qfz(self):
        if getattr(self, "_Qfz", None) is None:
            self._Qfz = self.prob.mesh.getInterpolationMat(self.rxLoc, "Fz")
        return self._Qfz

    def projectFields(self, u):
        """
            This function projects the fields onto the data space.

            First we project our B on to data location

            .. math::

                \\mathbf{B}_{rec} = \\mathbf{P} \\mathbf{B}

            then we take the dot product between B and b_0

            .. math ::

                \\text{TMI} = \\vec{B}_s \\cdot \\hat{B}_0

        """
        # TODO: There can be some different tyes of data like |B| or B

        gfx = self.Qfx * u["G"]
        gfy = self.Qfy * u["G"]
        gfz = self.Qfz * u["G"]

        fields = {"gx": gfx, "gy": gfy, "gz": gfz}
        return fields


class SrcField(Survey.BaseSrc):
    """ Define the inducing field """

    param = None  #: Inducing field param (Amp, Incl, Decl)

    def __init__(self, rxList, **kwargs):
        super().__init__(rxList, **kwargs)


class RxObs(Survey.BaseRx):
    """A station location must have be located in 3-D"""

    def __init__(self, locsXYZ, **kwargs):
        locs = locsXYZ
        assert locsXYZ.shape[1] == 3, "locs must in 3-D (x,y,z)."
        super().__init__(locs, "tmi", storeProjections=False, **kwargs)

    @property
    def nD(self):
        """Number of data in the receiver."""
        return self.locs[0].shape[0]


class BaseGravMap(Maps.IdentityMap):
    """BaseGravMap"""

    def __init__(self, mesh, **kwargs):
        Maps.IdentityMap.__init__(self, mesh)

    def _transform(self, m):

        return m

    def deriv(self, m):

        return sp.identity(self.nP)



import numpy as np
import scipy.sparse as sp
from six import integer_types
import warnings
import dask
import dask.array as da
from . import Utils
from .Tests import checkDerivative
from . import Maps
from . import Props


__all__ = ["BaseObjectiveFunction", "ComboObjectiveFunction", "L2ObjectiveFunction"]


class BaseObjectiveFunction(Props.BaseSimPEG):
    """
    Base Objective Function

    Inherit this to build your own objective function. If building a
    regularization, have a look at
    :class:`SimPEG.Regularization.BaseRegularization` as there are additional
    methods and properties tailored to regularization of a model. Similarly,
    for building a data misfit, see :class:`SimPEG.DataMisfit.BaseDataMisfit`.
    """

    counter = None
    debug = False

    mapPair = Maps.IdentityMap  #: Base class of expected maps
    _mapping = None  #: An IdentityMap instance.
    _hasFields = False  #: should we have the option to store fields

    _nP = None  #: number of parameters

    def __init__(self, nP=None, **kwargs):
        if nP is not None:
            self._nP = nP
        Utils.setKwargs(self, **kwargs)

    def __call__(self, x, f=None):
        raise NotImplementedError(
            "__call__ has not been implemented for {} yet".format(
                self.__class__.__name__
            )
        )

    @property
    def nP(self):
        """
        Number of model parameters expected.
        """
        if self._nP is not None:
            return self._nP
        if getattr(self, "mapping", None) is not None:
            return self.mapping.nP
        return "*"

    @property
    def _nC_residual(self):
        """
        Shape of the residual
        """
        if getattr(self, "mapping", None) is not None:
            return self.mapping.shape[0]
        else:
            return self.nP

    @property
    def mapping(self):
        """
        A `SimPEG.Maps` instance
        """
        if getattr(self, "_mapping") is None:
            if getattr(self, "_nP") is not None:
                self._mapping = self.mapPair(nP=self.nP)
            else:
                self._mapping = self.mapPair()
        return self._mapping

    @mapping.setter
    def mapping(self, value):
        assert isinstance(value, self.mapPair), (
            "mapping must be an instance of a {}, not a {}"
        ).format(self.mapPair, value.__class__.__name__)
        self._mapping = value

    @Utils.timeIt
    def __call__(self, x, f=None):
        raise NotImplementedError(
            "The method __call__ has not been implemented for {}".format(
                self.__class__.__name__
            )
        )

    @Utils.timeIt
    def deriv(self, x, **kwargs):
        """
        First derivative of the objective function with respect to the model
        """
        raise NotImplementedError(
            "The method deriv has not been implemented for {}".format(
                self.__class__.__name__
            )
        )

    @Utils.timeIt
    def deriv2(self, x, v=None, **kwargs):
        """
        Second derivative of the objective function with respect to the model
        """
        raise NotImplementedError(
            "The method _deriv2 has not been implemented for {}".format(
                self.__class__.__name__
            )
        )

    def _test_deriv(self, x=None, num=4, plotIt=False, **kwargs):
        print(f"Testing {self.__class__.__name__!s} Deriv")
        if x is None:
            if self.nP == "*":
                x = np.random.randn(np.random.randint(1e2, high=1e3))
            else:
                x = np.random.randn(self.nP)

        return checkDerivative(
            lambda m: [self(m), self.deriv(m)], x, num=num, plotIt=plotIt, **kwargs
        )

    def _test_deriv2(self, x=None, num=4, plotIt=False, **kwargs):
        print(f"Testing {self.__class__.__name__!s} Deriv2")
        if x is None:
            if self.nP == "*":
                x = np.random.randn(np.random.randint(1e2, high=1e3))
            else:
                x = np.random.randn(self.nP)

        v = x + 0.1 * np.random.rand(len(x))
        return checkDerivative(
            lambda m: [self.deriv(m).dot(v), self.deriv2(m, v=v)],
            x,
            num=num,
            expectedOrder=1,
            plotIt=plotIt,
            **kwargs,
        )

    def test(self, x=None, num=4, plotIt=False, **kwargs):
        """
        Run a convergence test on both the first and second derivatives - they
        should be second order!
        """
        deriv = self._test_deriv(x=x, num=num, **kwargs)
        deriv2 = self._test_deriv2(x=x, num=num, plotIt=False, **kwargs)
        return deriv & deriv2

    __numpy_ufunc__ = True

    def __add__(self, objfct2):

        if isinstance(objfct2, Utils.Zero):
            return self

        if not isinstance(objfct2, BaseObjectiveFunction):
            raise Exception(
                "Cannot add type {} to an objective function. Only "
                "ObjectiveFunctions can be added together".format(
                    objfct2.__class__.__name__
                )
            )

        if (
            self.__class__.__name__ != "ComboObjectiveFunction"
        ):  # not isinstance(self, ComboObjectiveFunction):
            self = 1 * self

        if (
            objfct2.__class__.__name__ != "ComboObjectiveFunction"
        ):  # not isinstance(objfct2, ComboObjectiveFunction):
            objfct2 = 1 * objfct2

        objfctlist = self.objfcts + objfct2.objfcts
        multipliers = self.multipliers + objfct2.multipliers

        return ComboObjectiveFunction(objfcts=objfctlist, multipliers=multipliers)

    def __radd__(self, objfct2):
        return self + objfct2

    def __mul__(self, multiplier):
        return ComboObjectiveFunction([self], [multiplier])

    def __rmul__(self, multiplier):
        return self * multiplier

    def __div__(self, denominator):
        return self.__mul__(1.0 / denominator)

    def __truediv__(self, denominator):
        return self.__mul__(1.0 / denominator)

    def __rdiv__(self, denominator):
        return self.__mul__(1.0 / denominator)


class ComboObjectiveFunction(BaseObjectiveFunction):
    """
    A composite objective function that consists of multiple objective
    functions. Objective functions are stored in a list, and multipliers
    are stored in a parallel list.

    .. code::python

        import SimPEG.ObjectiveFunction
        phi1 = ObjectiveFunction.L2ObjectiveFunction(nP=10)
        phi2 = ObjectiveFunction.L2ObjectiveFunction(nP=10)

        phi = 2*phi1 + 3*phi2

    is equivalent to

        .. code::python

            import SimPEG.ObjectiveFunction
            phi1 = ObjectiveFunction.L2ObjectiveFunction(nP=10)
            phi2 = ObjectiveFunction.L2ObjectiveFunction(nP=10)

            phi = ObjectiveFunction.ComboObjectiveFunction(
                [phi1, phi2], [2, 3]
            )

    """

    _multiplier_types = (float, None, Utils.Zero, np.float64,) + (int,)  # Directive
    _multipliers = None

    def __init__(self, objfcts=[], multipliers=None, **kwargs):

        if multipliers is None:
            multipliers = len(objfcts) * [1]

        self._nP = "*"

        assert len(objfcts) == len(multipliers), (
            "Must have the same number of Objective Functions and Multipliers "
            "not {} and {}".format(len(objfcts), len(multipliers))
        )

        def validate_list(objfctlist, multipliers):
            """
            ensure that the number of parameters expected by each objective
            function is the same, ensure that if multpliers are supplied, that
            list matches the length of the objective function list
            """
            for fct, mult in zip(objfctlist, multipliers):
                assert isinstance(fct, BaseObjectiveFunction), (
                    "Unrecognized objective function type {} in objfcts. "
                    "All entries in objfcts must inherit from "
                    "ObjectiveFunction".format(fct.__class__.__name__)
                )

                assert type(mult) in self._multiplier_types, (
                    "Objective Functions can only be multiplied by a "
                    "float, or a properties.Float, not a {}, {}".format(
                        type(mult), mult
                    )
                )

                if fct.nP != "*":
                    if self._nP != "*":
                        assert self._nP == fct.nP, (
                            "Objective Functions must all have the same "
                            "nP={}, not {}".format(self.nP, [f.nP for f in objfcts])
                        )
                    else:
                        self._nP = fct.nP

        validate_list(objfcts, multipliers)

        self.objfcts = objfcts
        self._multipliers = multipliers

        super().__init__(**kwargs)

    def __len__(self):
        return len(self.multipliers)

    def __getitem__(self, key):
        return self.multipliers[key], self.objfcts[key]

    @property
    def __len__(self):
        return self.objfcts.__len__

    @property
    def multipliers(self):
        return self._multipliers

    @multipliers.setter
    def multipliers(self, value):
        for val in value:
            assert (
                type(val) in self._multiplier_types
            ), "Multiplier must be in type {} not {}".format(
                self._multiplier_types, type(val)
            )

        assert len(value) == len(self.objfcts), (
            "the length of multipliers should be the same as the number of"
            " objective functions ({}), not {}".format(len(self.objfcts, len(value)))
        )

        self._multipliers = value

    def __call__(self, m, f=None):

        fct = []
        for i, phi in enumerate(self):
            multiplier, objfct = phi
            if multiplier == 0.0:  # don't evaluate the fct
                continue
            else:
                if f is not None and objfct._hasFields:
                    fct += [multiplier * objfct(m, f=f[i])]
                else:
                    fct += [multiplier * objfct(m)]

        stack = da.vstack(fct)

        return da.sum(stack, axis=0).compute()

    def deriv(self, m, f=None):
        """
        First derivative of the composite objective function is the sum of the
        derivatives of each objective function in the list, weighted by their
        respective multplier.

        :param numpy.ndarray m: model
        :param SimPEG.Fields f: Fields object (if applicable)
        """

        # @dask.delayed
        # def rowSum(arr):
        #     sumIt = 0
        #     for i in range(len(arr)):
        #         sumIt += arr[i]
        #     return sumIt

        g = []
        for i, phi in enumerate(self):
            multiplier, objfct = phi
            if multiplier == 0.0:  # don't evaluate the fct
                continue
            else:
                if f is not None and objfct._hasFields:
                    g += [multiplier * objfct.deriv(m, f=f[i])]
                else:
                    g += [multiplier * objfct.deriv(m)]

        stack = da.vstack(g)

        return da.sum(stack, axis=0).compute()

    def deriv2(self, m, v=None, f=None):
        """
        Second derivative of the composite objective function is the sum of the
        second derivatives of each objective function in the list, weighted by
        their respective multplier.

        :param numpy.ndarray m: model
        :param numpy.ndarray v: vector we are multiplying by
        :param SimPEG.Fields f: Fields object (if applicable)
        """
        # @dask.delayed
        # def rowSum(arr):
        #     sumIt = 0
        #     for i in range(len(arr)):
        #         sumIt += arr[i]
        #     return sumIt

        H = []
        for i, phi in enumerate(self):
            multiplier, objfct = phi
            if multiplier == 0.0:  # don't evaluate the fct
                continue
            else:
                if f is not None and objfct._hasFields:

                    H += [multiplier * objfct.deriv2(m, v, f=f[i])]
                else:
                    H += [multiplier * objfct.deriv2(m, v)]

        if isinstance(H[0], dask.array.Array):

            stack = da.vstack(H)

            return da.sum(stack, axis=0).compute()

        else:
            sumIt = 0
            for i in range(len(H)):
                sumIt += H[i]
            return sumIt

    # This assumes all objective functions have a W.
    # The base class currently does not.
    @property
    def W(self):
        """
        W matrix for the full objective function. Includes multiplying by the
        square root of alpha.
        """
        W = []
        for mult, fct in self:
            curW = np.sqrt(mult) * fct.W
            if not isinstance(curW, Utils.Zero):
                W.append(curW)
        return sp.vstack(W)


class L2ObjectiveFunction(BaseObjectiveFunction):
    """
    An L2-Objective Function

    .. math::

        \\phi = \frac{1}{2}||\\mathbf{W} \\mathbf{m}||^2
    """

    def __init__(self, W=None, **kwargs):

        super().__init__(**kwargs)
        if W is not None:
            if self.nP == "*":
                self._nP = W.shape[1]
        self._W = W

    @property
    def W(self):
        """
        Weighting matrix. The default if not sepcified is an identity.
        """
        if getattr(self, "_W", None) is None:
            if self._nC_residual != "*":
                self._W = sp.eye(self._nC_residual)
            else:
                self._W = Utils.Identity()
        return self._W

    def __call__(self, m):
        r = self.W * (self.mapping * m)
        return 0.5 * r.dot(r)

    def deriv(self, m):
        """
        First derivative with respect to the model

        :param numpy.ndarray m: model
        """
        return self.mapping.deriv(m).T * (self.W.T * (self.W * (self.mapping * m)))

    def deriv2(self, m, v=None):
        """
        Second derivative with respect to the model

        :param numpy.ndarray m: model
        :param numpy.ndarray v: vector to multiply by
        """
        if v is not None:
            return self.mapping.deriv(m).T * (
                self.W.T * (self.W * (self.mapping.deriv(m) * v))
            )
        W = self.W * self.mapping.deriv(m)
        return W.T * W



from .matutils import mkvc, ndgrid
import numpy as np
from scipy.interpolate import griddata, interp1d, interp2d, NearestNDInterpolator
import numpy as np
import scipy.sparse as sp
from scipy.spatial import Delaunay
from scipy.sparse.linalg import bicgstab
from scipy.spatial import cKDTree
from scipy.interpolate import LinearNDInterpolator
import discretize as Mesh
from discretize.utils import closestPoints, kron3, speye


def surface2ind_topo(mesh, topo, gridLoc="CC", method="linear", fill_value=np.nan):
    """
    Get active indices from topography

    Parameters
    ----------

    :param TensorMesh mesh: TensorMesh object on which to discretize topography
    :param numpy.ndarray topo: [X,Y,Z] topographic data
    :param str gridLoc: 'CC' or 'N'. Default is 'CC'.
                        Discretize the topography
                        on cells-center 'CC' or nodes 'N'
    :param str method: 'nearest' or 'linear' or 'cubic'. Default is 'nearest'.
                       Interpolation method for the topographic data
    :param float fill_value: default is np.nan. Filling value for extrapolation

    Returns
    -------

    :param numpy.array actind: index vector for the active cells on the mesh
                               below the topography
    """

    assert gridLoc in [
        "N",
        "CC",
    ], "Value of gridLoc must be 'N' (nodal) or 'CC' (cell center)"

    dim = mesh.dim - 1

    if mesh.dim == 3:
        assert topo.shape[1] == 3, "topo locations of shape (*, 3) required for 3D mesh"
        if method == "linear":
            tri2D = Delaunay(topo[:, :2])
            z_interpolate = LinearNDInterpolator(tri2D, topo[:, 2])
        else:
            z_interpolate = NearestNDInterpolator(topo[:, :2], topo[:, 2])
    elif mesh.dim == 2:
        assert topo.shape[1] == 2, "topo locations of shape (*, 2) required for 2D mesh"
        z_interpolate = interp1d(
            topo[:, 0], topo[:, 1], bounds_error=False, fill_value=np.nan, kind=method
        )
    else:
        assert topo.ndim == 1, "topo locations of shape (*, ) required for 1D mesh"

    if gridLoc == "CC":
        locations = mesh.gridCC

        if mesh.dim == 1:
            active = np.zeros(mesh.nC, dtype="bool")
            active[np.searchsorted(mesh.vectorCCx, topo).max() :] = True
            return active

    elif gridLoc == "N":

        if mesh.dim == 3:
            locations = np.vstack(
                [
                    mesh.gridCC
                    + (np.c_[-1, 1, 1][:, None] * mesh.h_gridded / 2.0).squeeze(),
                    mesh.gridCC
                    + (np.c_[-1, -1, 1][:, None] * mesh.h_gridded / 2.0).squeeze(),
                    mesh.gridCC
                    + (np.c_[1, 1, 1][:, None] * mesh.h_gridded / 2.0).squeeze(),
                    mesh.gridCC
                    + (np.c_[1, -1, 1][:, None] * mesh.h_gridded / 2.0).squeeze(),
                ]
            )

        elif mesh.dim == 2:
            locations = np.vstack(
                [
                    mesh.gridCC
                    + (np.c_[-1, 1][:, None] * mesh.h_gridded / 2.0).squeeze(),
                    mesh.gridCC
                    + (np.c_[1, 1][:, None] * mesh.h_gridded / 2.0).squeeze(),
                ]
            )

        else:
            active = np.zeros(mesh.nC, dtype="bool")
            active[np.searchsorted(mesh.vectorNx, topo).max() :] = True

            return active

    # Interpolate z values on CC or N
    z_topo = z_interpolate(locations[:, :-1]).squeeze()

    # Apply nearest neighbour if in extrapolation
    ind_nan = np.isnan(z_topo)

    if np.any(ind_nan):
        tree = cKDTree(topo)
        _, ind = tree.query(locations[ind_nan, :])
        z_topo[ind_nan] = topo[ind, dim]

    # Create an active bool of all True
    active = np.all(
        (locations[:, dim] < z_topo).reshape((mesh.nC, -1), order="F"), axis=1
    )

    return active.ravel()


def tileSurveyPoints(locs, nRefine, minimize=True, method="cluster"):
    """
        Function to tile an survey points into smaller square subsets of points

        :param numpy.ndarray locs: n x 2 array of locations [x,y]
        :param integer nRefine: number of tiles (for 'cluster'), or number of
            refinement steps ('other')
        :param Bool minimize: shrink tile sizes to minimum
        :param string method: set to 'cluster' to use better quality clustering, or anything
            else to use more memory efficient method for large problems

        RETURNS:
        :param numpy.ndarray: Return a list of arrays with the for the SW and NE
                            limits of each tiles
        :param integer binCount: Number of points in each tile
        :param numpy.array labels: Cluster index of each point n=0:(nTargetTiles-1)
        :param numpy.array tile_numbers: Vector of tile numbers for each count in binCount

        NOTE: All X Y and xy products are legacy now values, and are only used
        for plotting functions. They are not used in any calculations and could
        be dropped from the return calls in future versions.


    """

    if method == "cluster":
        # Best for smaller problems
        from sklearn.cluster import AgglomerativeClustering

        # Cluster
        cluster = AgglomerativeClustering(
            n_clusters=nRefine, affinity="euclidean", linkage="ward"
        )
        cluster.fit_predict(locs[:, :2])

        # nData in each tile
        binCount = np.zeros(int(nRefine))

        # x and y limits on each tile
        X1 = np.zeros_like(binCount)
        X2 = np.zeros_like(binCount)
        Y1 = np.zeros_like(binCount)
        Y2 = np.zeros_like(binCount)

        for ii in range(int(nRefine)):

            mask = cluster.labels_ == ii
            X1[ii] = locs[mask, 0].min()
            X2[ii] = locs[mask, 0].max()
            Y1[ii] = locs[mask, 1].min()
            Y2[ii] = locs[mask, 1].max()
            binCount[ii] = mask.sum()

        xy1 = np.c_[X1[binCount > 0], Y1[binCount > 0]]
        xy2 = np.c_[X2[binCount > 0], Y2[binCount > 0]]

        # Get the tile numbers that exist, for compatibility with the next method
        tile_numbers = np.unique(cluster.labels_)
        return [xy1, xy2], binCount[binCount > 0], cluster.labels_, tile_numbers

    else:
        # Works on larger problems
        # Initialize variables
        # Test each refinement level for maximum space coverage
        nTx = 1
        nTy = 1
        for ii in range(int(nRefine + 1)):

            nTx += 1
            nTy += 1

            testx = np.percentile(locs[:, 0], np.arange(0, 100, 100 / nTx))
            testy = np.percentile(locs[:, 1], np.arange(0, 100, 100 / nTy))

            # if ii > 0:
            dx = testx[:-1] - testx[1:]
            dy = testy[:-1] - testy[1:]

            if np.mean(dx) > np.mean(dy):
                nTx -= 1
            else:
                nTy -= 1

            print(nTx, nTy)
        tilex = np.percentile(locs[:, 0], np.arange(0, 100, 100 / nTx))
        tiley = np.percentile(locs[:, 1], np.arange(0, 100, 100 / nTy))

        X1, Y1 = np.meshgrid(tilex, tiley)
        X2, Y2 = np.meshgrid(
            np.r_[tilex[1:], locs[:, 0].max()], np.r_[tiley[1:], locs[:, 1].max()]
        )

        # Plot data and tiles
        X1, Y1, X2, Y2 = mkvc(X1), mkvc(Y1), mkvc(X2), mkvc(Y2)
        binCount = np.zeros_like(X1)
        tile_labels = np.zeros_like(locs[:, 0])
        for ii in range(X1.shape[0]):

            mask = (
                (locs[:, 0] >= X1[ii])
                * (locs[:, 0] <= X2[ii])
                * (locs[:, 1] >= Y1[ii])
                * (locs[:, 1] <= Y2[ii])
            ) == 1

            # Re-adjust the window size for tight fit
            if minimize:

                if mask.sum():
                    X1[ii], X2[ii] = locs[:, 0][mask].min(), locs[:, 0][mask].max()
                    Y1[ii], Y2[ii] = locs[:, 1][mask].min(), locs[:, 1][mask].max()

            tile_labels[mask] = ii
            binCount[ii] = mask.sum()

        xy1 = np.c_[X1[binCount > 0], Y1[binCount > 0]]
        xy2 = np.c_[X2[binCount > 0], Y2[binCount > 0]]

        # Get the tile numbers that exist
        # Since some tiles may have 0 data locations, and are removed by
        # [binCount > 0], the tile numbers are no longer contiguous 0:nTiles
        tile_numbers = np.unique(tile_labels)
        return [xy1, xy2], binCount[binCount > 0], tile_labels, tile_numbers


def minCurvatureInterp(
    locs,
    data,
    mesh=None,
    vectorX=None,
    vectorY=None,
    vectorZ=None,
    gridSize=10,
    tol=1e-5,
    iterMax=200,
    method="spline",
):
    """
    Interpolate properties with a minimum curvature interpolation
    :param locs:  numpy.array of size n-by-3 of point locations
    :param data: numpy.array of size n-by-m of values to be interpolated
    :param vectorX: numpy.ndarray Gridded locations along x-axis [Default:None]
    :param vectorY: numpy.ndarray Gridded locations along y-axis [Default:None]
    :param vectorZ: numpy.ndarray Gridded locations along z-axis [Default:None]
    :param gridSize: numpy float Grid point seperation in meters [DEFAULT:10]
    :param method: 'relaxation' || 'spline' [Default]
    :param tol: float tol=1e-5 [Default] Convergence criteria
    :param iterMax: int iterMax=None [Default] Maximum number of iterations

    :return: numpy.array of size nC-by-m of interpolated values

    """

    def av_extrap(n):
        """Define 1D averaging operator from cell-centers to nodes."""
        Av = sp.spdiags(
            (0.5 * np.ones((n, 1)) * [1, 1]).T, [-1, 0], n + 1, n, format="csr"
        )
        Av[0, 1], Av[-1, -2] = 0.5, 0.5
        return Av

    def aveCC2F(grid):
        "Construct the averaging operator on cell cell centers to faces."
        if grid.ndim == 1:
            aveCC2F = av_extrap(grid.shape[0])
        elif grid.ndim == 2:
            aveCC2F = sp.vstack(
                (
                    sp.kron(speye(grid.shape[1]), av_extrap(grid.shape[0])),
                    sp.kron(av_extrap(grid.shape[1]), speye(grid.shape[0])),
                ),
                format="csr",
            )
        elif grid.ndim == 3:
            aveCC2F = sp.vstack(
                (
                    kron3(
                        speye(grid.shape[2]),
                        speye(grid.shape[1]),
                        av_extrap(grid.shape[0]),
                    ),
                    kron3(
                        speye(grid.shape[2]),
                        av_extrap(grid.shape[1]),
                        speye(grid.shape[0]),
                    ),
                    kron3(
                        av_extrap(grid.shape[2]),
                        speye(grid.shape[1]),
                        speye(grid.shape[0]),
                    ),
                ),
                format="csr",
            )
        return aveCC2F

    assert locs.shape[0] == data.shape[0], (
        "Number of interpolated locs " + "must match number of data"
    )

    if vectorY is not None:
        assert locs.shape[1] >= 2, (
            "Found vectorY as an input."
            + " Point locations must contain X and Y coordinates."
        )

    if vectorZ is not None:
        assert locs.shape[1] == 3, (
            "Found vectorZ as an input."
            + " Point locations must contain X, Y and Z coordinates."
        )

    ndim = locs.shape[1]

    # Define a new grid based on data extent
    if mesh is None:
        if vectorX is None:
            xmin, xmax = locs[:, 0].min(), locs[:, 0].max()
            nCx = int((xmax - xmin) / gridSize)
            vectorX = xmin + np.cumsum(np.ones(nCx) * gridSize)

        if vectorY is None and ndim >= 2:
            ymin, ymax = locs[:, 1].min(), locs[:, 1].max()
            nCy = int((ymax - ymin) / gridSize)
            vectorY = ymin + np.cumsum(np.ones(nCy) * gridSize)

        if vectorZ is None and ndim == 3:
            zmin, zmax = locs[:, 2].min(), locs[:, 2].max()
            nCz = int((zmax - zmin) / gridSize)
            vectorZ = zmin + np.cumsum(np.ones(nCz) * gridSize)

        if ndim == 3:

            mesh = Mesh.TensorMesh(
                [
                    np.ones(nCx) * gridSize,
                    np.ones(nCy) * gridSize,
                    np.ones(nCz) * gridSize,
                ]
            )
            mesh.x0 = [xmin, ymin, zmin]

            # gridCy, gridCx, gridCz = np.meshgrid(vectorY, vectorX, vectorZ)
            # gridCC = np.c_[mkvc(gridCx), mkvc(gridCy), mkvc(gridCz)]
        elif ndim == 2:

            mesh = Mesh.TensorMesh([np.ones(nCx) * gridSize, np.ones(nCy) * gridSize])
            mesh.x0 = [xmin, ymin]
            # gridCy, gridCx = np.meshgrid(vectorY, vectorX)
            # gridCC = np.c_[mkvc(gridCx), mkvc(gridCy)]
        else:
            mesh = Mesh.TensorMesh([np.ones(nCx) * gridSize])
            mesh.x0 = xmin
    else:

        gridSize = mesh.h_gridded.min()

    gridCC = mesh.gridCC
    # Build the cKDTree for distance lookup
    tree = cKDTree(locs)
    # Get the grid location
    d, ind = tree.query(gridCC, k=1)

    if data.ndim == 1:
        data = np.c_[data]

    if method == "relaxation":
        # Ave = aveCC2F(gridCC)
        Ave = mesh.aveCC2F

        count = 0
        residual = 1.0

        m = np.zeros((gridCC.shape[0], data.shape[1]))
        # Begin with neighrest primers
        for ii in range(m.shape[1]):
            # F = NearestNDInterpolator(mesh.gridCC[ijk], data[:, ii])
            m[:, ii] = data[ind, ii]

        while np.all([count < iterMax, residual > tol]):
            for ii in range(m.shape[1]):
                # Reset the closest cell grid to the contraints
                m[d < gridSize, ii] = data[ind[d < gridSize], ii]
            mtemp = m
            m = Ave.T * (Ave * m)
            residual = np.linalg.norm(m - mtemp) / np.linalg.norm(mtemp)
            count += 1

    elif method == "spline":

        ndat = locs.shape[0]
        # nC = int(nCx*nCy)

        A = np.zeros((ndat, ndat))
        for i in range(ndat):

            r = (locs[i, 0] - locs[:, 0]) ** 2.0 + (locs[i, 1] - locs[:, 1]) ** 2.0
            A[i, :] = r.T * (np.log((r.T + 1e-8) ** 0.5) - 1.0)

        # Compute new solution
        nC = gridCC.shape[0]
        m = np.zeros((nC, data.shape[1]))

        # Solve system for the weights
        for dd in range(data.shape[1]):
            w = bicgstab(A, data[:, dd], tol=tol)

            # We can parallelize this part later
            for i in range(nC):

                r = (gridCC[i, 0] - locs[:, 0]) ** 2.0 + (
                    gridCC[i, 1] - locs[:, 1]
                ) ** 2.0
                m[i, dd] = np.sum(w[0] * r.T * (np.log((r.T + 1e-8) ** 0.5) - 1.0))

    return mesh, m


def activeTopoLayer(mesh, topo, index=0):
    """
        Find the ith layer below topo
    """

    actv = np.zeros(mesh.nC, dtype="bool")
    # Get cdkTree to find top layer
    tree = cKDTree(mesh.gridCC)

    def ismember(a, b):
        bind = {}
        for i, elt in enumerate(b):
            if elt not in bind:
                bind[elt] = i
        return np.vstack([bind.get(itm, None) for itm in a])

    grid_x, grid_y = np.meshgrid(mesh.vectorCCx, mesh.vectorCCy)
    zInterp = mkvc(
        griddata(topo[:, :2], topo[:, 2], (grid_x, grid_y), method="nearest")
    )
    r, inds = tree.query(np.c_[mkvc(grid_x), mkvc(grid_y), zInterp])
    inds = np.unique(inds)

    # Extract neighbors from operators
    # Dz = mesh._cellGradzStencil
    # Iz, Jz, _ = sp.find(Dz)
    # jz = np.sort(Jz[np.argsort(Iz)].reshape((int(Iz.shape[0]/2), 2)), axis=1)
    # for ii in range(index):

    #     members = ismember(inds, jz[:, 1])
    #     inds = np.squeeze(jz[members, 0])

    # actv[inds] = True

    max_level = mesh.max_level

    for layer in range(index):

        nn = []
        for ind in inds.tolist():

            skip = 2 ** (max_level - mesh[int(ind)]._level)
            if (np.floor(layer / skip) % 2) > 0:
                nn += [mesh[int(ind)].neighbors[4]]
            else:
                nn += [int(ind)]

        inds = np.hstack(nn)

    actv[np.hstack(nn)] = True

    return actv


def transfer_to_mesh(mesh_in, mesh_out, values, return_indices=False):
    """
        transfer_to_mesh(mesh_input, mesh_output, values, return_indices=False)

        Function to transfer values from an input mesh to another.

        Parameters
        ----------
        mesh_in: discetize.mesh
            Mesh object with input values

        mesh_out: discetize.mesh
            Mesh object to be transfered onto

        values: ndarray of floats (mesh_in.nC, nD)
            Cell center values to be transfered

        return_indices: bool optional
            Option to return an array of int (mesh_out.nC, ) for the indices
            of `mesh_in` such that:
            `a = values[indices]` are values tranfered from `mesh_in` inside
            `mesh_out`

        Returns
        -------

        values_out: ndarray or float (mesh_out.nC, nD)
            Values transfered from mesh_in to mesh_out

        incides: ndarray of int (mesh_out.nC, )
            Integer values from mesh_in to mesh_out

    """

    tree = cKDTree(mesh_in.gridCC)

    _, cell_indices = tree.query(mesh_out.gridCC)

    if values.ndim > 1:
        values_out = values[cell_indices, :]
    else:
        values_out = values[cell_indices]

    if return_indices:
        return (values_out, cell_indices)

    return values_out



from . import (
    simpegEM1D,
    simpegPF,
    selection,
    utils,
    processing,
    simulation,
    export,
    plotting,
)

# from .plotting import plot_plan_data_selection
# from .selection import ObjectDataSelection



import numpy as np

from .. import Utils, Mesh
from SimPEG.EM.Static import DC


def plot_pseudoSection(
    DCsurvey,
    axs,
    surveyType="dipole-dipole",
    dataType="appConductivity",
    clim=None,
    scale="linear",
    sameratio=True,
    pcolorOpts={},
):
    """
        Read list of 2D tx-rx location and plot a speudo-section of apparent
        resistivity.

        Assumes flat topo for now...

        Input:
        :param d2D, z0
        :switch surveyType -> Either 'pole-dipole' | 'dipole-dipole'
        :switch dataType=-> Either 'appResistivity' | 'appConductivity' | 'volt' (potential)
        :scale -> Either 'linear' (default) | 'log'
        Output:
        :figure scatter plot overlayed on image

        Edited Feb 17th, 2016

        @author: dominiquef

    """
    from scipy.interpolate import griddata
    import pylab as plt

    # Set depth to 0 for now
    z0 = 0.0

    # Pre-allocate
    midx = []
    midz = []
    rho = []
    LEG = []
    count = 0  # Counter for data

    for ii in range(DCsurvey.nSrc):

        Tx = DCsurvey.srcList[ii].loc
        Rx = DCsurvey.srcList[ii].rxList[0].locs

        nD = DCsurvey.srcList[ii].rxList[0].nD

        data = DCsurvey.dobs[count : count + nD]
        count += nD

        # Get distances between each poles A-B-M-N
        if surveyType == "pole-dipole":

            MA = np.abs(Tx[0] - Rx[0][:, 0])
            NA = np.abs(Tx[0] - Rx[1][:, 0])
            MN = np.abs(Rx[1][:, 0] - Rx[0][:, 0])

            # Create mid-point location
            Cmid = Tx[0]
            Pmid = (Rx[0][:, 0] + Rx[1][:, 0]) / 2
            # if DCsurvey.mesh.dim == 2:
            #     zsrc = Tx[1]
            # elif DCsurvey.mesh.dim ==3:
            zsrc = Tx[2]

        elif surveyType == "dipole-dipole":
            MA = np.abs(Tx[0][0] - Rx[0][:, 0])
            MB = np.abs(Tx[1][0] - Rx[0][:, 0])
            NA = np.abs(Tx[0][0] - Rx[1][:, 0])
            NB = np.abs(Tx[1][0] - Rx[1][:, 0])

            # Create mid-point location
            Cmid = (Tx[0][0] + Tx[1][0]) / 2
            Pmid = (Rx[0][:, 0] + Rx[1][:, 0]) / 2
            # if DCsurvey.mesh.dim == 2:
            #     zsrc = (Tx[0][1] + Tx[1][1])/2
            # elif DCsurvey.mesh.dim ==3:
            zsrc = (Tx[0][2] + Tx[1][2]) / 2

        # Change output for dataType
        if surveyType == "pole-dipole":

            leg = data * 2 * np.pi * MA * (MA + MN) / MN

        elif surveyType == "dipole-dipole":

            leg = data * 2 * np.pi / (1 / MA - 1 / MB + 1 / NB - 1 / NA)
            LEG.append(1.0 / (2 * np.pi) * (1 / MA - 1 / MB + 1 / NB - 1 / NA))
        else:
            print(""" dataType must be 'pole-dipole' | 'dipole-dipole' """)
            break

        if dataType == "volt":
            if scale == "linear":
                rho = np.hstack([rho, data])
            elif scale == "log":
                rho = np.hstack([rho, np.log10(abs(data))])

        else:

            # Compute pant leg of apparent rho

            if dataType == "appConductivity":

                leg = abs(1.0 / leg)

            elif dataType == "appResistivity":

                leg = abs(leg)

            else:
                print(
                    """dataType must be 'appResistivity' | 'appConductivity' | 'volt' """
                )
                break

            if scale == "linear":
                rho = np.hstack([rho, leg])
            elif scale == "log":
                rho = np.hstack([rho, np.log10(leg)])

        midx = np.hstack([midx, (Cmid + Pmid) / 2])
        # if DCsurvey.mesh.dim==3:
        midz = np.hstack([midz, -np.abs(Cmid - Pmid) / 2 + zsrc])
        # elif DCsurvey.mesh.dim==2:
        #     midz = np.hstack([midz, -np.abs(Cmid-Pmid)/2 + zsrc ])
    ax = axs

    # Grid points
    grid_x, grid_z = np.mgrid[np.min(midx) : np.max(midx), np.min(midz) : np.max(midz)]

    grid_rho = griddata(np.c_[midx, midz], rho.T, (grid_x, grid_z), method="linear")

    if clim is None:
        vmin, vmax = rho.min(), rho.max()
    else:
        vmin, vmax = clim[0], clim[1]

    grid_rho = np.ma.masked_where(np.isnan(grid_rho), grid_rho)
    ph = plt.pcolormesh(
        grid_x[:, 0],
        grid_z[0, :],
        grid_rho.T,
        clim=(vmin, vmax),
        vmin=vmin,
        vmax=vmax,
        **pcolorOpts
    )

    if scale == "log":
        cbar = plt.colorbar(
            format="$10^{%.1f}$", fraction=0.04, orientation="horizontal"
        )
    elif scale == "linear":
        cbar = plt.colorbar(format="%.1f", fraction=0.04, orientation="horizontal")

    if dataType == "appConductivity":
        cbar.set_label("App.Cond", size=12)

    elif dataType == "appResistivity":
        cbar.set_label("App.Res.", size=12)

    elif dataType == "volt":
        cbar.set_label("Potential (V)", size=12)

    cmin, cmax = cbar.get_clim()
    ticks = np.linspace(cmin, cmax, 3)
    cbar.set_ticks(ticks)
    cbar.ax.tick_params(labelsize=10)

    # Plot apparent resistivity
    ax.plot(midx, midz, "k.", ms=1)

    if sameratio:
        plt.gca().set_aspect("equal", adjustable="box")

    return ph, ax, cbar, LEG


def gen_DCIPsurvey(endl, mesh, surveyType, a, b, n):
    """
        Load in endpoints and survey specifications to generate Tx, Rx location
        stations.

        Assumes flat topo for now...

        Input:
        :param endl -> input endpoints [x1, y1, z1, x2, y2, z2]
        :object mesh -> SimPEG mesh object
        :switch surveyType -> "dipole-dipole" (dipole-dipole) | "pole-dipole" (pole-dipole) | 'gradient'
        : param a, n -> pole seperation, number of rx dipoles per tx

        Output:
        :param Tx, Rx -> List objects for each tx location
            Lines: P1x, P1y, P1z, P2x, P2y, P2z

        Created on Wed December 9th, 2015

        @author: dominiquef
        !! Require clean up to deal with DCsurvey
    """

    def xy_2_r(x1, x2, y1, y2):
        r = np.sqrt(np.sum((x2 - x1) ** 2 + (y2 - y1) ** 2))
        return r

    # Evenly distribute electrodes and put on surface
    # Mesure survey length and direction
    dl_len = xy_2_r(endl[0, 0], endl[1, 0], endl[0, 1], endl[1, 1])

    dl_x = (endl[1, 0] - endl[0, 0]) / dl_len
    dl_y = (endl[1, 1] - endl[0, 1]) / dl_len

    nstn = int(np.floor(dl_len / a))

    # Compute discrete pole location along line
    stn_x = endl[0, 0] + np.array(range(int(nstn))) * dl_x * a
    stn_y = endl[0, 1] + np.array(range(int(nstn))) * dl_y * a

    if mesh.dim == 2:
        ztop = mesh.vectorNy[-1]
        # Create line of P1 locations
        M = np.c_[stn_x, np.ones(nstn).T * ztop]
        # Create line of P2 locations
        N = np.c_[stn_x + a * dl_x, np.ones(nstn).T * ztop]

    elif mesh.dim == 3:
        ztop = mesh.vectorNz[-1]
        # Create line of P1 locations
        M = np.c_[stn_x, stn_y, np.ones(nstn).T * ztop]
        # Create line of P2 locations
        N = np.c_[stn_x + a * dl_x, stn_y + a * dl_y, np.ones(nstn).T * ztop]

    # Build list of Tx-Rx locations depending on survey type
    # Dipole-dipole: Moving tx with [a] spacing -> [AB a MN1 a MN2 ... a MNn]
    # Pole-dipole: Moving pole on one end -> [A a MN1 a MN2 ... MNn a B]
    SrcList = []

    if surveyType != "gradient":

        for ii in range(0, int(nstn) - 1):

            if surveyType == "dipole-dipole":
                tx = np.c_[M[ii, :], N[ii, :]]
            elif surveyType == "pole-dipole":
                tx = np.c_[M[ii, :], M[ii, :]]
            else:
                raise Exception(
                    'The surveyType must be "dipole-dipole" or "pole-dipole"'
                )

            # Rx.append(np.c_[M[ii+1:indx, :], N[ii+1:indx, :]])

            # Current elctrode seperation
            AB = xy_2_r(tx[0, 1], endl[1, 0], tx[1, 1], endl[1, 1])

            # Number of receivers to fit
            nstn = int(np.min([np.floor((AB - b) / a), n]))

            # Check if there is enough space, else break the loop
            if nstn <= 0:
                continue

            # Compute discrete pole location along line
            stn_x = N[ii, 0] + dl_x * b + np.array(range(int(nstn))) * dl_x * a
            stn_y = N[ii, 1] + dl_y * b + np.array(range(int(nstn))) * dl_y * a

            # Create receiver poles

            if mesh.dim == 3:
                # Create line of P1 locations
                P1 = np.c_[stn_x, stn_y, np.ones(nstn).T * ztop]
                # Create line of P2 locations
                P2 = np.c_[stn_x + a * dl_x, stn_y + a * dl_y, np.ones(nstn).T * ztop]
                rxClass = DC.Rx.Dipole(P1, P2)

            elif mesh.dim == 2:
                # Create line of P1 locations
                P1 = np.c_[stn_x, np.ones(nstn).T * ztop]
                # Create line of P2 locations
                P2 = np.c_[stn_x + a * dl_x, np.ones(nstn).T * ztop]
                rxClass = DC.Rx.Dipole_ky(P1, P2)

            if surveyType == "dipole-dipole":
                srcClass = DC.Src.Dipole([rxClass], M[ii, :], N[ii, :])
            elif surveyType == "pole-dipole":
                srcClass = DC.Src.Pole([rxClass], M[ii, :])
            SrcList.append(srcClass)

    elif surveyType == "gradient":

        # Gradient survey takes the "b" parameter to define the limits of a
        # square survey grid. The pole seperation within the receiver grid is
        # define the "a" parameter.

        # Get the edge limit of survey area
        min_x = endl[0, 0] + dl_x * b
        min_y = endl[0, 1] + dl_y * b

        max_x = endl[1, 0] - dl_x * b
        max_y = endl[1, 1] - dl_y * b

        # Define the size of the survey grid (square for now)
        box_l = np.sqrt((min_x - max_x) ** 2 + (min_y - max_y) ** 2)
        box_w = box_l / 2.0

        nstn = int(np.floor(box_l / a))

        # Compute discrete pole location along line
        stn_x = min_x + np.array(range(int(nstn))) * dl_x * a
        stn_y = min_y + np.array(range(int(nstn))) * dl_y * a

        # Define number of cross lines
        nlin = int(np.floor(box_w / a))
        lind = range(-nlin, nlin + 1)

        npoles = int(nstn * len(lind))

        rx = np.zeros([npoles, 6])
        for ii in range(len(lind)):

            # Move station location to current survey line This is a
            # perpendicular move then line survey orientation, hence the y, x
            # switch
            lxx = stn_x - lind[ii] * a * dl_y
            lyy = stn_y + lind[ii] * a * dl_x

            M = np.c_[lxx, lyy, np.ones(nstn).T * ztop]
            N = np.c_[lxx + a * dl_x, lyy + a * dl_y, np.ones(nstn).T * ztop]
            rx[(ii * nstn) : ((ii + 1) * nstn), :] = np.c_[M, N]

            if mesh.dim == 3:
                rxClass = DC.Rx.Dipole(rx[:, :3], rx[:, 3:])
            elif mesh.dim == 2:
                M = M[:, [0, 2]]
                N = N[:, [0, 2]]
                rxClass = DC.Rx.Dipole_ky(rx[:, [0, 2]], rx[:, [3, 5]])
            srcClass = DC.Src.Dipole([rxClass], (endl[0, :]), (endl[1, :]))
        SrcList.append(srcClass)
    else:
        print(
            """surveyType must be either 'pole-dipole', 'dipole-dipole' or 'gradient'. """
        )

    survey = DC.Survey(SrcList)

    return survey


def writeUBC_DCobs(fileName, DCsurvey, dim, formatType, iptype=0):
    """
        Write UBC GIF DCIP 2D or 3D observation file

        :param string fileName: including path where the file is written out
        :param Survey DCsurvey: DC survey class object
        :param string dim:  either '2D' | '3D'
        :param string surveyType:  either 'SURFACE' | 'GENERAL'
        :rtype: file
        :return: UBC2D-Data file
    """

    assert (dim == "2D") | (dim == "3D"), "Data must be either '2D' | '3D'"

    assert (
        (formatType == "SURFACE") | (formatType == "GENERAL") | (formatType == "SIMPLE")
    ), "Data must be either 'SURFACE' | 'GENERAL' | 'SIMPLE'"

    fid = open(fileName, "w")

    if iptype != 0:
        fid.write("IPTYPE=%i\n" % iptype)

    else:
        fid.write("! " + formatType + " FORMAT\n")

    count = 0

    for ii in range(DCsurvey.nSrc):

        tx = np.c_[DCsurvey.srcList[ii].loc]

        if np.shape(tx)[0] == 3:
            surveyType = "pole-dipole"

        else:
            surveyType = "dipole-dipole"

        rx = DCsurvey.srcList[ii].rxList[0].locs

        nD = DCsurvey.srcList[ii].nD

        M = rx[0]
        N = rx[1]

        # Adapt source-receiver location for dim and surveyType
        if dim == "2D":

            if formatType == "SIMPLE":

                # fid.writelines("%e " % ii for ii in Utils.mkvc(tx[0, :]))
                A = np.repeat(tx[0, 0], M.shape[0], axis=0)

                if surveyType == "pole-dipole":
                    B = np.repeat(tx[0, 0], M.shape[0], axis=0)

                else:
                    B = np.repeat(tx[1, 0], M.shape[0], axis=0)

                M = M[:, 0]
                N = N[:, 0]

                np.savetxt(
                    fid,
                    np.c_[
                        A,
                        B,
                        M,
                        N,
                        DCsurvey.dobs[count : count + nD],
                        DCsurvey.std[count : count + nD],
                    ],
                    delimiter=" ",
                    newline="\n",
                )

            else:

                if formatType == "SURFACE":

                    fid.writelines("%f " % ii for ii in Utils.mkvc(tx[0, :]))
                    M = M[:, 0]
                    N = N[:, 0]

                if formatType == "GENERAL":

                    # Flip sign for z-elevation to depth
                    tx[2::2, :] = -tx[2::2, :]

                    fid.writelines("%e " % ii for ii in Utils.mkvc(tx[::2, :]))
                    M = M[:, 0::2]
                    N = N[:, 0::2]

                    # Flip sign for z-elevation to depth
                    M[:, 1::2] = -M[:, 1::2]
                    N[:, 1::2] = -N[:, 1::2]

                fid.write("%i\n" % nD)
                np.savetxt(
                    fid,
                    np.c_[
                        M,
                        N,
                        DCsurvey.dobs[count : count + nD],
                        DCsurvey.std[count : count + nD],
                    ],
                    delimiter=" ",
                    newline="\n",
                )

        if dim == "3D":

            if formatType == "SURFACE":

                fid.writelines("%e " % ii for ii in Utils.mkvc(tx[:, 0:2].T))
                M = M[:, 0:2]
                N = N[:, 0:2]

            if formatType == "GENERAL":

                fid.writelines("%e " % ii for ii in Utils.mkvc(tx.T))

            fid.write("%i\n" % nD)
            np.savetxt(
                fid,
                np.c_[
                    M,
                    N,
                    DCsurvey.dobs[count : count + nD],
                    DCsurvey.std[count : count + nD],
                ],
                fmt="%e",
                delimiter=" ",
                newline="\n",
            )
            fid.write("\n")

        count += nD

    fid.close()


def convertObs_DC3D_to_2D(survey, lineID, flag="local"):
    """
        Read DC survey and projects the coordinate system
        according to the flag = 'Xloc' | 'Yloc' | 'local' (default)
        In the 'local' system, station coordinates are referenced
        to distance from the first srcLoc[0].loc[0]

        The Z value is preserved, but Y coordinates zeroed.

        Input:
        :param survey3D

        Output:
        :figure survey2D

        Edited April 6th, 2016

        @author: dominiquef

    """

    def stn_id(v0, v1, r):
        """
        Compute station ID along line
        """

        dl = int(v0.dot(v1)) * r

        return dl

    def r_unit(p1, p2):
        """
        r_unit(x, y) : Function computes the unit vector
        between two points with coordinates p1(x1, y1) and p2(x2, y2)

        """

        assert len(p1) == len(p2), "locs must be the same shape."

        dx = []
        for ii in range(len(p1)):
            dx.append(p2[ii] - p1[ii])

        # Compute length of vector
        r = np.linalg.norm(np.asarray(dx))

        if r != 0:
            vec = dx / r

        else:
            vec = np.zeros(len(p1))

        return vec, r

    srcList2D = []

    srcMat = getSrc_locs(survey)

    # Find all unique line id
    uniqueID = np.unique(lineID)

    for jj in range(len(uniqueID)):

        indx = np.where(lineID == uniqueID[jj])[0]

        # Find origin of survey
        r = 1e8  # Initialize to some large number

        Tx = srcMat[indx]

        if np.all(Tx[0:3] == Tx[3:]):
            surveyType = "pole-dipole"

        else:
            surveyType = "dipole-dipole"

        x0 = Tx[0][0:2]  # Define station zero along line

        vecTx, r1 = r_unit(x0, Tx[-1][0:2])

        for ii in range(len(indx)):

            # Get all receivers
            Rx = survey.srcList[indx[ii]].rxList[0].locs
            nrx = Rx[0].shape[0]

            if flag == "local":
                # Find A electrode along line
                vec, r = r_unit(x0, Tx[ii][0:2])
                A = stn_id(vecTx, vec, r)

                if surveyType != "pole-dipole":
                    # Find B electrode along line
                    vec, r = r_unit(x0, Tx[ii][3:5])
                    B = stn_id(vecTx, vec, r)

                M = np.zeros(nrx)
                N = np.zeros(nrx)
                for kk in range(nrx):

                    # Find all M electrodes along line
                    vec, r = r_unit(x0, Rx[0][kk, 0:2])
                    M[kk] = stn_id(vecTx, vec, r)

                    # Find all N electrodes along line
                    vec, r = r_unit(x0, Rx[1][kk, 0:2])
                    N[kk] = stn_id(vecTx, vec, r)
            elif flag == "Yloc":
                """ Flip the XY axis locs"""
                A = Tx[ii][1]

                if surveyType != "pole-dipole":
                    B = Tx[ii][4]

                M = Rx[0][:, 1]
                N = Rx[1][:, 1]

            elif flag == "Xloc":
                """ Copy the rx-tx locs"""
                A = Tx[ii][0]

                if surveyType != "pole-dipole":
                    B = Tx[ii][3]

                M = Rx[0][:, 0]
                N = Rx[1][:, 0]

            rxClass = DC.Rx.Dipole(
                np.c_[M, np.zeros(nrx), Rx[0][:, 2]],
                np.c_[N, np.zeros(nrx), Rx[1][:, 2]],
            )

            if surveyType == "pole-dipole":
                srcList2D.append(DC.Src.Pole([rxClass], np.asarray([A, 0, Tx[ii][2]])))

            elif surveyType == "dipole-dipole":
                srcList2D.append(
                    DC.Src.Dipole(
                        [rxClass], np.r_[A, 0, Tx[ii][2]], np.r_[B, 0, Tx[ii][5]]
                    )
                )

    survey2D = DC.SurveyDC.Survey(srcList2D)
    survey2D.dobs = survey.dobs
    survey2D.std = survey.std

    return survey2D


def readUBC_DC2DModel(fileName):
    """
        Read UBC GIF 2DTensor model and generate 2D Tensor model in simpeg

        Input:
        :param fileName, path to the UBC GIF 2D model file

        Output:
        :param SimPEG TensorMesh 2D object
        :return

        Created on Thu Nov 12 13:14:10 2015

        @author: dominiquef

    """

    # Open fileand skip header... assume that we know the mesh already
    obsfile = np.genfromtxt(fileName, delimiter=" \n", dtype=np.str, comments="!")

    dim = np.array(obsfile[0].split(), dtype=float)

    temp = np.array(obsfile[1].split(), dtype=float)

    if len(temp) > 1:
        model = np.zeros(dim)

        for ii in range(len(obsfile) - 1):
            mm = np.array(obsfile[ii + 1].split(), dtype=float)
            model[:, ii] = mm

        model = model[:, ::-1]

    else:

        if len(obsfile[1:]) == 1:
            mm = np.array(obsfile[1:].split(), dtype=float)

        else:
            mm = np.array(obsfile[1:], dtype=float)

        # Permute the second dimension to flip the order
        model = mm.reshape(dim[1], dim[0])

        model = model[::-1, :]
        model = np.transpose(model, (1, 0))

    model = Utils.mkvc(model)

    return model


def readUBC_DC2Dpre(fileName):
    """
        Read UBC GIF DCIP 2D observation file and generate arrays for tx-rx location

        Input:
        :param fileName, path to the UBC GIF 3D obs file

        Output:
        DCsurvey
        :return

        Created on Mon March 9th, 2016 << Doug's 70th Birthday !! >>

        @author: dominiquef

    """

    # Load file
    obsfile = np.genfromtxt(fileName, delimiter=" \n", dtype=np.str, comments="!")

    # Pre-allocate
    srcLists = []
    Rx = []
    d = []
    zflag = True  # Flag for z value provided

    for ii in range(obsfile.shape[0]):

        if not obsfile[ii]:
            continue

        # First line is transmitter with number of receivers

        temp = np.fromstring(obsfile[ii], dtype=float, sep=" ").T

        # Check if z value is provided, if False -> nan
        if len(temp) == 5:
            tx = np.r_[temp[0], np.nan, np.nan, temp[1], np.nan, np.nan]
            zflag = False

        else:
            tx = np.r_[temp[0], np.nan, temp[1], temp[2], np.nan, temp[3]]

        if zflag:
            rx = np.c_[temp[4], np.nan, temp[5], temp[6], np.nan, temp[7]]

        else:
            rx = np.c_[temp[2], np.nan, np.nan, temp[3], np.nan, np.nan]
            # Check if there is data with the location

        d.append(temp[-1])

        Rx = DC.Rx.Dipole(rx[:, :3], rx[:, 3:])
        srcLists.append(DC.Src.Dipole([Rx], tx[:3], tx[3:]))

    # Create survey class
    survey = DC.SurveyDC.Survey(srcLists)

    survey.dobs = np.asarray(d)

    return {"DCsurvey": survey}


def readUBC_DC3Dobs(fileName):
    """
        Read UBC GIF DCIP 3D observation file and generate arrays for tx-rx location

        Input:
        :param fileName, path to the UBC GIF 3D obs file

        Output:
        :param rx, tx, d, wd
        :return

        Created on Mon December 7th, 2015

        @author: dominiquef

    """

    # Load file
    obsfile = np.genfromtxt(fileName, delimiter=" \n", dtype=np.str, comments="!")

    # Pre-allocate
    srcLists = []
    Rx = []
    d = []
    wd = []
    zflag = True  # Flag for z value provided

    # Countdown for number of obs/tx
    count = 0
    for ii in range(obsfile.shape[0]):

        if not obsfile[ii]:
            continue

        # First line is transmitter with number of receivers
        if count == 0:
            rx = []
            temp = np.fromstring(obsfile[ii], dtype=float, sep=" ").T
            count = int(temp[-1])

            # Check if z value is provided, if False -> nan
            if len(temp) == 5:
                tx = np.r_[temp[0:2], np.nan, temp[0:2], np.nan]
                zflag = False

            else:
                tx = temp[:-1]

            continue

        temp = np.fromstring(obsfile[ii], dtype=float, sep=" ")

        if zflag:

            rx.append(temp[:-2])
            # Check if there is data with the location
            if len(temp) == 8:
                d.append(temp[-2])
                wd.append(temp[-1])

        else:
            rx.append(np.r_[temp[0:2], np.nan, temp[0:2], np.nan])

            # Check if there is data with the location
            if len(temp) == 6:
                d.append(temp[-2])
                wd.append(temp[-1])

        count = count - 1

        # Reach the end of transmitter block
        if count == 0:
            rx = np.asarray(rx)
            Rx = DC.Rx.Dipole(rx[:, :3], rx[:, 3:])
            srcLists.append(DC.Src.Dipole([Rx], tx[:3], tx[3:]))

    survey = DC.SurveyDC.Survey(srcLists)
    survey.dobs = np.asarray(d)
    survey.std = np.asarray(wd)

    return {"DCsurvey": survey}


def xy_2_lineID(DCsurvey):
    """
        Read DC survey class and append line ID.
        Assumes that the locations are listed in the order
        they were collected. May need to generalize for random
        point locations, but will be more expensive

        Input:
        :param DCdict Vectors of station location

        Output:
        :param LineID Vector of integers
        :return

        Created on Thu Feb 11, 2015

        @author: dominiquef

    """

    # Compute unit vector between two points
    nstn = DCsurvey.nSrc

    # Pre-allocate space
    lineID = np.zeros(nstn)

    linenum = 0
    indx = 0

    for ii in range(nstn):

        if ii == 0:

            A = DCsurvey.srcList[ii].loc[0]
            B = DCsurvey.srcList[ii].loc[1]

            xout = np.mean([A[0:2], B[0:2]], axis=0)

            xy0 = A[:2]
            xym = xout

            # Deal with replicate pole location
            if np.all(xy0 == xym):

                xym[0] = xym[0] + 1e-3

            continue

        A = DCsurvey.srcList[ii].loc[0]
        B = DCsurvey.srcList[ii].loc[1]

        xin = np.mean([A[0:2], B[0:2]], axis=0)

        # Compute vector between neighbours
        vec1, r1 = r_unit(xout, xin)

        # Compute vector between current stn and mid-point
        vec2, r2 = r_unit(xym, xin)

        # Compute vector between current stn and start line
        vec3, r3 = r_unit(xy0, xin)

        # Compute vector between mid-point and start line
        vec4, r4 = r_unit(xym, xy0)

        # Compute dot product
        ang1 = np.abs(vec1.dot(vec2))
        ang2 = np.abs(vec3.dot(vec4))

        # If the angles are smaller then 45d, than next point is on a new line
        if ((ang1 < np.cos(np.pi / 4.0)) | (ang2 < np.cos(np.pi / 4.0))) & (
            np.all(np.r_[r1, r2, r3, r4] > 0)
        ):

            # Re-initiate start and mid-point location
            xy0 = A[:2]
            xym = xin

            # Deal with replicate pole location
            if np.all(xy0 == xym):

                xym[0] = xym[0] + 1e-3

            linenum += 1
            indx = ii

        else:
            xym = np.mean([xy0, xin], axis=0)

        lineID[ii] = linenum
        xout = xin

    return lineID


def r_unit(p1, p2):
    """
    r_unit(x, y) : Function computes the unit vector
    between two points with coordinates p1(x1, y1) and p2(x2, y2)

    """

    assert len(p1) == len(p2), "locs must be the same shape."

    dx = []
    for ii in range(len(p1)):
        dx.append(p2[ii] - p1[ii])

    # Compute length of vector
    r = np.linalg.norm(np.asarray(dx))

    if r != 0:
        vec = dx / r

    else:
        vec = np.zeros(len(p1))

    return vec, r


def getSrc_locs(survey):
    """
        Read in a DC survey class and extract the xyz location of all
        sources.

        :param DCclass survey: Input Static.DC class
        :return numpy.array srcLocs: Locations of sources

    """

    srcMat = []

    for src in survey.srcList:

        srcMat.append(np.hstack(src.loc))

    srcMat = np.vstack(srcMat)

    return srcMat


def gettopoCC(mesh, actind):
    """
        Get topography from active indices of mesh.

    """

    if mesh.dim == 3:

        mesh2D = Mesh.TensorMesh([mesh.hx, mesh.hy], mesh.x0[:2])
        zc = mesh.gridCC[:, 2]
        ACTIND = actind.reshape((mesh.vnC[0] * mesh.vnC[1], mesh.vnC[2]), order="F")
        ZC = zc.reshape((mesh.vnC[0] * mesh.vnC[1], mesh.vnC[2]), order="F")
        topo = np.zeros(ZC.shape[0])
        topoCC = np.zeros(ZC.shape[0])
        for i in range(ZC.shape[0]):
            ind = np.argmax(ZC[i, :][ACTIND[i, :]])
            topo[i] = ZC[i, :][ACTIND[i, :]].max() + mesh.hz[ACTIND[i, :]][ind] * 0.5
            topoCC[i] = ZC[i, :][ACTIND[i, :]].max()

        return mesh2D, topoCC

    elif mesh.dim == 2:

        mesh1D = Mesh.TensorMesh([mesh.hx], [mesh.x0[0]])
        yc = mesh.gridCC[:, 1]
        ACTIND = actind.reshape((mesh.vnC[0], mesh.vnC[1]), order="F")
        YC = yc.reshape((mesh.vnC[0], mesh.vnC[1]), order="F")
        topo = np.zeros(YC.shape[0])
        topoCC = np.zeros(YC.shape[0])
        for i in range(YC.shape[0]):
            ind = np.argmax(YC[i, :][ACTIND[i, :]])
            topo[i] = YC[i, :][ACTIND[i, :]].max() + mesh.hy[ACTIND[i, :]][ind] * 0.5
            topoCC[i] = YC[i, :][ACTIND[i, :]].max()

        return mesh1D, topoCC


def drapeTopotoLoc(mesh, pts, actind=None, topo=None):
    """
        Drape location right below (cell center) the topography
    """
    if mesh.dim == 2:
        if pts.ndim > 1:
            raise Exception("pts should be 1d array")
    elif mesh.dim == 3:
        if pts.shape[1] == 3:
            raise Exception("shape of pts should be (x,3)")
    else:
        raise NotImplementedError()
    if actind is None:
        actind = Utils.surface2ind_topo(mesh, topo)

    meshtemp, topoCC = gettopoCC(mesh, actind)
    inds = Utils.closestPoints(meshtemp, pts)
    out = np.c_[pts, topoCC[inds]]
    return out



import numpy as np
from scipy.constants import mu_0, pi
from scipy.special import erf
from .. import Utils


def hzAnalyticDipoleF(r, freq, sigma, secondary=True, mu=mu_0):
    """
    4.56 in Ward and Hohmann

    .. plot::

        import matplotlib.pyplot as plt
        from .. import EM
        freq = np.logspace(-1, 6, 61)
        test = EM.Analytics.FDEM.hzAnalyticDipoleF(100, freq, 0.001, secondary=False)
        plt.loglog(freq, abs(test.real))
        plt.loglog(freq, abs(test.imag))
        plt.title('Response at $r$=100m')
        plt.xlabel('Frequency')
        plt.ylabel('Response')
        plt.legend(('real','imag'))
        plt.show()

    """
    r = np.abs(r)
    k = np.sqrt(-1j * 2.0 * np.pi * freq * mu * sigma)

    m = 1
    front = m / (2.0 * np.pi * (k ** 2) * (r ** 5))
    back = 9 - (
        9 + 9j * k * r - 4 * (k ** 2) * (r ** 2) - 1j * (k ** 3) * (r ** 3)
    ) * np.exp(-1j * k * r)
    hz = front * back

    if secondary:
        hp = -1 / (4 * np.pi * r ** 3)
        hz = hz - hp

    if hz.ndim == 1:
        hz = Utils.mkvc(hz, 2)

    return hz


def MagneticDipoleWholeSpace(XYZ, srcLoc, sig, f, moment=1.0, orientation="X", mu=mu_0):
    """
    Analytical solution for a dipole in a whole-space.

    Equation 2.57 of Ward and Hohmann

    TODOs:
        - set it up to instead take a mesh & survey
        - add E-fields
        - handle multiple frequencies
        - add divide by zero safety


    .. plot::

        from .. import EM
        import matplotlib.pyplot as plt
        from scipy.constants import mu_0
        freqs = np.logspace(-2,5,100)
        Bx, By, Bz = EM.Analytics.FDEM.MagneticDipoleWholeSpace([0,100,0], [0,0,0], 1e-2, freqs, moment=1, orientation='Z')
        plt.loglog(freqs, np.abs(Bz.real)/mu_0, 'b')
        plt.loglog(freqs, np.abs(Bz.imag)/mu_0, 'r')
        plt.legend(('real','imag'))
        plt.show()


    """

    if not isinstance(orientation, str):
        if np.allclose(orientation, np.r_[1.0, 0.0, 0.0]):
            orientation = "X"
        elif np.allclose(orientation, np.r_[0.0, 1.0, 0.0]):
            orientation = "Y"
        elif np.allclose(orientation, np.r_[0.0, 0.0, 1.0]):
            orientation = "Z"
        else:
            raise NotImplementedError("arbitrary orientations not implemented")

    XYZ = Utils.asArray_N_x_Dim(XYZ, 3)

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx ** 2.0 + dy ** 2.0 + dz ** 2.0)
    k = np.sqrt(-1j * 2.0 * np.pi * f * mu * sig)
    kr = k * r

    front = moment / (4.0 * pi * r ** 3.0) * np.exp(-1j * kr)
    mid = -(kr ** 2.0) + 3.0 * 1j * kr + 3.0

    if orientation.upper() == "X":
        Hx = front * ((dx / r) ** 2.0 * mid + (kr ** 2.0 - 1j * kr - 1.0))
        Hy = front * ((dx * dy / r ** 2.0) * mid)
        Hz = front * ((dx * dz / r ** 2.0) * mid)

    elif orientation.upper() == "Y":
        Hx = front * ((dy * dx / r ** 2.0) * mid)
        Hy = front * ((dy / r) ** 2.0 * mid + (kr ** 2.0 - 1j * kr - 1.0))
        Hz = front * ((dy * dz / r ** 2.0) * mid)

    elif orientation.upper() == "Z":
        Hx = front * ((dx * dz / r ** 2.0) * mid)
        Hy = front * ((dy * dz / r ** 2.0) * mid)
        Hz = front * ((dz / r) ** 2.0 * mid + (kr ** 2.0 - 1j * kr - 1.0))

    Bx = mu * Hx
    By = mu * Hy
    Bz = mu * Hz

    if Bx.ndim == 1:
        Bx = Utils.mkvc(Bx, 2)

    if By.ndim == 1:
        By = Utils.mkvc(By, 2)

    if Bz.ndim == 1:
        Bz = Utils.mkvc(Bz, 2)

    return Bx, By, Bz


def ElectricDipoleWholeSpace(
    XYZ, srcLoc, sig, f, current=1.0, length=1.0, orientation="X", mu=mu_0
):
    XYZ = Utils.asArray_N_x_Dim(XYZ, 3)

    dx = XYZ[:, 0] - srcLoc[0]
    dy = XYZ[:, 1] - srcLoc[1]
    dz = XYZ[:, 2] - srcLoc[2]

    r = np.sqrt(dx ** 2.0 + dy ** 2.0 + dz ** 2.0)
    k = np.sqrt(-1j * 2.0 * np.pi * f * mu * sig)
    kr = k * r

    front = current * length / (4.0 * np.pi * sig * r ** 3) * np.exp(-1j * k * r)
    mid = -(k ** 2) * r ** 2 + 3 * 1j * k * r + 3

    # Ex = front*((dx**2 / r**2)*mid + (k**2 * r**2 -1j*k*r))
    # Ey = front*(dx*dy  / r**2)*mid
    # Ez = front*(dx*dz  / r**2)*mid

    if orientation.upper() == "X":
        Ex = front * ((dx ** 2 / r ** 2) * mid + (k ** 2 * r ** 2 - 1j * k * r - 1.0))
        Ey = front * (dx * dy / r ** 2) * mid
        Ez = front * (dx * dz / r ** 2) * mid
        return Ex, Ey, Ez

    elif orientation.upper() == "Y":
        #  x--> y, y--> z, z-->x
        Ey = front * ((dy ** 2 / r ** 2) * mid + (k ** 2 * r ** 2 - 1j * k * r - 1.0))
        Ez = front * (dy * dz / r ** 2) * mid
        Ex = front * (dy * dx / r ** 2) * mid
        return Ex, Ey, Ez

    elif orientation.upper() == "Z":
        # x --> z, y --> x, z --> y
        Ez = front * ((dz ** 2 / r ** 2) * mid + (k ** 2 * r ** 2 - 1j * k * r - 1.0))
        Ex = front * (dz * dx / r ** 2) * mid
        Ey = front * (dz * dy / r ** 2) * mid
        return Ex, Ey, Ez
        # return Ey, Ez, Ex



import scipy.sparse as sp
import numpy as np
from .. import Problem, Utils, Solver as SimpegSolver
from SimPEG.EM.Base import BaseEMProblem
from SimPEG.EM.TDEM.SurveyTDEM import Survey as SurveyTDEM
from SimPEG.EM.TDEM.FieldsTDEM import (
    FieldsTDEM,
    Fields3D_b,
    Fields3D_e,
    Fields3D_h,
    Fields3D_j,
    Fields_Derivs,
)
from scipy.constants import mu_0
import time


class BaseTDEMProblem(Problem.BaseTimeProblem, BaseEMProblem):
    """
    We start with the first order form of Maxwell's equations, eliminate and
    solve the second order form. For the time discretization, we use backward
    Euler.
    """

    surveyPair = SurveyTDEM  #: A SimPEG.EM.TDEM.SurveyTDEM Class
    fieldsPair = FieldsTDEM  #: A SimPEG.EM.TDEM.FieldsTDEM Class

    def __init__(self, mesh, **kwargs):
        BaseEMProblem.__init__(self, mesh, **kwargs)

    # def fields_nostore(self, m):
    #     """
    #     Solve the forward problem without storing fields

    #     :param numpy.array m: inversion model (nP,)
    #     :rtype: numpy.array
    #     :return numpy.array: numpy.array (nD,)

    #     """

    def fields(self, m):
        """
        Solve the forward problem for the fields.

        :param numpy.array m: inversion model (nP,)
        :rtype: SimPEG.EM.TDEM.FieldsTDEM
        :return F: fields object
        """

        tic = time.time()
        self.model = m

        F = self.fieldsPair(self.mesh, self.survey)

        # set initial fields
        F[:, self._fieldType + "Solution", 0] = self.getInitialFields()

        # timestep to solve forward
        if self.verbose:
            print("{}\nCalculating fields(m)\n{}".format("*" * 50, "*" * 50))
        Ainv = None
        for tInd, dt in enumerate(self.timeSteps):
            # keep factors if dt is the same as previous step b/c A will be the
            # same
            if Ainv is not None and (tInd > 0 and dt != self.timeSteps[tInd - 1]):
                Ainv.clean()
                Ainv = None

            if Ainv is None:
                A = self.getAdiag(tInd)
                if self.verbose:
                    print(f"Factoring...   (dt = {dt:e})")
                Ainv = self.Solver(A, **self.solverOpts)
                if self.verbose:
                    print("Done")

            rhs = self.getRHS(tInd + 1)  # this is on the nodes of the time mesh
            Asubdiag = self.getAsubdiag(tInd)

            if self.verbose:
                print("    Solving...   (tInd = {:d})".format(tInd + 1))
            # taking a step
            sol = Ainv * (rhs - Asubdiag * F[:, (self._fieldType + "Solution"), tInd])

            if self.verbose:
                print("    Done...")

            if sol.ndim == 1:
                sol.shape = (sol.size, 1)
            F[:, self._fieldType + "Solution", tInd + 1] = sol
        if self.verbose:
            print("{}\nDone calculating fields(m)\n{}".format("*" * 50, "*" * 50))
        Ainv.clean()
        return F

    def Jvec(self, m, v, f=None):
        """
        Jvec computes the sensitivity times a vector

        .. math::
            \\mathbf{J} \\mathbf{v} = \\frac{d\\mathbf{P}}{d\\mathbf{F}} \\left(
            \\frac{d\\mathbf{F}}{d\\mathbf{u}} \\frac{d\\mathbf{u}}{d\\mathbf{m}} +
            \\frac{\\partial\\mathbf{F}}{\\partial\\mathbf{m}} \\right) \\mathbf{v}

        where

        .. math::
            \\mathbf{A} \\frac{d\\mathbf{u}}{d\\mathbf{m}} +
            \\frac{\\partial \\mathbf{A}(\\mathbf{u}, \\mathbf{m})}
            {\\partial\\mathbf{m}} =
            \\frac{d \\mathbf{RHS}}{d \\mathbf{m}}
        """

        if f is None:
            f = self.fields(m)

        ftype = self._fieldType + "Solution"  # the thing we solved for
        self.model = m

        # mat to store previous time-step's solution deriv times a vector for
        # each source
        # size: nu x nSrc

        # this is a bit silly

        # if self._fieldType is 'b' or self._fieldType is 'j':
        #     ifields = np.zeros((self.mesh.nF, len(Srcs)))
        # elif self._fieldType is 'e' or self._fieldType is 'h':
        #     ifields = np.zeros((self.mesh.nE, len(Srcs)))

        # for i, src in enumerate(self.survey.srcList):
        dun_dm_v = np.hstack(
            [
                Utils.mkvc(self.getInitialFieldsDeriv(src, v, f=f), 2)
                for src in self.survey.srcList
            ]
        )
        # can over-write this at each timestep
        # store the field derivs we need to project to calc full deriv
        df_dm_v = Fields_Derivs(self.mesh, self.survey)

        Adiaginv = None

        for tInd, dt in zip(range(self.nT), self.timeSteps):
            # keep factors if dt is the same as previous step b/c A will be the
            # same
            if Adiaginv is not None and (tInd > 0 and dt != self.timeSteps[tInd - 1]):
                Adiaginv.clean()
                Adiaginv = None

            if Adiaginv is None:
                A = self.getAdiag(tInd)
                Adiaginv = self.Solver(A, **self.solverOpts)

            Asubdiag = self.getAsubdiag(tInd)

            for i, src in enumerate(self.survey.srcList):

                # here, we are lagging by a timestep, so filling in as we go
                for projField in {rx.projField for rx in src.rxList}:
                    df_dmFun = getattr(f, "_%sDeriv" % projField, None)
                    # df_dm_v is dense, but we only need the times at
                    # (rx.P.T * ones > 0)
                    # This should be called rx.footprint

                    df_dm_v[src, f"{projField}Deriv", tInd] = df_dmFun(
                        tInd, src, dun_dm_v[:, i], v
                    )

                un_src = f[src, ftype, tInd + 1]

                # cell centered on time mesh
                dA_dm_v = self.getAdiagDeriv(tInd, un_src, v)
                # on nodes of time mesh
                dRHS_dm_v = self.getRHSDeriv(tInd + 1, src, v)

                dAsubdiag_dm_v = self.getAsubdiagDeriv(tInd, f[src, ftype, tInd], v)

                JRHS = dRHS_dm_v - dAsubdiag_dm_v - dA_dm_v

                # step in time and overwrite
                if tInd != len(self.timeSteps + 1):
                    dun_dm_v[:, i] = Adiaginv * (JRHS - Asubdiag * dun_dm_v[:, i])

        Jv = []
        for src in self.survey.srcList:
            for rx in src.rxList:
                Jv.append(
                    rx.evalDeriv(
                        src,
                        self.mesh,
                        self.timeMesh,
                        f,
                        Utils.mkvc(df_dm_v[src, "%sDeriv" % rx.projField, :]),
                    )
                )
        Adiaginv.clean()
        # del df_dm_v, dun_dm_v, Asubdiag
        # return Utils.mkvc(Jv)
        return np.hstack(Jv)

    def Jtvec(self, m, v, f=None):

        """
        Jvec computes the adjoint of the sensitivity times a vector

        .. math::
            \\mathbf{J}^\\top \\mathbf{v} =  \\left(
            \\frac{d\\mathbf{u}}{d\\mathbf{m}} ^ \\top
            \\frac{d\\mathbf{F}}{d\\mathbf{u}} ^ \\top  +
            \\frac{\\partial\\mathbf{F}}{\\partial\\mathbf{m}} ^ \\top \\right)
            \\frac{d\\mathbf{P}}{d\\mathbf{F}} ^ \\top \\mathbf{v}

        where

        .. math::
            \\frac{d\\mathbf{u}}{d\\mathbf{m}} ^\\top \\mathbf{A}^\\top  +
            \\frac{d\\mathbf{A}(\\mathbf{u})}{d\\mathbf{m}} ^ \\top =
            \\frac{d \\mathbf{RHS}}{d \\mathbf{m}} ^ \\top
        """

        if f is None:
            f = self.fields(m)

        self.model = m
        ftype = self._fieldType + "Solution"  # the thing we solved for

        # Ensure v is a data object.
        if not isinstance(v, self.dataPair):
            v = self.dataPair(self.survey, v)

        df_duT_v = Fields_Derivs(self.mesh, self.survey)

        # same size as fields at a single timestep
        ATinv_df_duT_v = np.zeros(
            (len(self.survey.srcList), len(f[self.survey.srcList[0], ftype, 0])),
            dtype=float,
        )
        JTv = np.zeros(m.shape, dtype=float)

        # Loop over sources and receivers to create a fields object:
        # PT_v, df_duT_v, df_dmT_v
        # initialize storage for PT_v (don't need to preserve over sources)
        PT_v = Fields_Derivs(self.mesh, self.survey)
        for src in self.survey.srcList:
            # Looping over initializing field class is appending memory!
            # PT_v = Fields_Derivs(self.mesh, self.survey) # initialize storage
            # #for PT_v (don't need to preserve over sources)
            # initialize size
            df_duT_v[src, f"{self._fieldType}Deriv", :] = np.zeros_like(
                f[src, self._fieldType, :]
            )

            for rx in src.rxList:
                PT_v[src, f"{rx.projField}Deriv", :] = rx.evalDeriv(
                    src,
                    self.mesh,
                    self.timeMesh,
                    f,
                    Utils.mkvc(v[src, rx]),
                    adjoint=True,
                )  # this is +=

                # PT_v = np.reshape(curPT_v,(len(curPT_v)/self.timeMesh.nN,
                # self.timeMesh.nN), order='F')
                df_duTFun = getattr(f, f"_{rx.projField}Deriv", None)

                for tInd in range(self.nT + 1):
                    cur = df_duTFun(
                        tInd,
                        src,
                        None,
                        Utils.mkvc(PT_v[src, f"{rx.projField}Deriv", tInd]),
                        adjoint=True,
                    )

                    df_duT_v[src, f"{self._fieldType}Deriv", tInd] = df_duT_v[
                        src, f"{self._fieldType}Deriv", tInd
                    ] + Utils.mkvc(cur[0], 2)
                    JTv = cur[1] + JTv

        del PT_v  # no longer need this

        AdiagTinv = None

        # Do the back-solve through time
        # if the previous timestep is the same: no need to refactor the matrix
        # for tInd, dt in zip(range(self.nT), self.timeSteps):

        for tInd in reversed(range(self.nT)):
            # tInd = tIndP - 1
            if AdiagTinv is not None and (
                tInd <= self.nT and self.timeSteps[tInd] != self.timeSteps[tInd + 1]
            ):
                AdiagTinv.clean()
                AdiagTinv = None

            # refactor if we need to
            if AdiagTinv is None:  # and tInd > -1:
                Adiag = self.getAdiag(tInd)
                AdiagTinv = self.Solver(Adiag.T, **self.solverOpts)

            if tInd < self.nT - 1:
                Asubdiag = self.getAsubdiag(tInd + 1)

            for isrc, src in enumerate(self.survey.srcList):

                # solve against df_duT_v
                if tInd >= self.nT - 1:
                    # last timestep (first to be solved)
                    ATinv_df_duT_v[isrc, :] = (
                        AdiagTinv * df_duT_v[src, f"{self._fieldType}Deriv", tInd + 1]
                    )
                elif tInd > -1:
                    ATinv_df_duT_v[isrc, :] = AdiagTinv * (
                        Utils.mkvc(df_duT_v[src, f"{self._fieldType}Deriv", tInd + 1])
                        - Asubdiag.T * Utils.mkvc(ATinv_df_duT_v[isrc, :])
                    )

                if tInd < self.nT:
                    dAsubdiagT_dm_v = self.getAsubdiagDeriv(
                        tInd, f[src, ftype, tInd], ATinv_df_duT_v[isrc, :], adjoint=True
                    )
                else:
                    dAsubdiagT_dm_v = Utils.Zero()

                dRHST_dm_v = self.getRHSDeriv(
                    tInd + 1, src, ATinv_df_duT_v[isrc, :], adjoint=True
                )  # on nodes of time mesh

                un_src = f[src, ftype, tInd + 1]
                # cell centered on time mesh
                dAT_dm_v = self.getAdiagDeriv(
                    tInd, un_src, ATinv_df_duT_v[isrc, :], adjoint=True
                )

                JTv = JTv + Utils.mkvc(-dAT_dm_v - dAsubdiagT_dm_v + dRHST_dm_v)

        # del df_duT_v, ATinv_df_duT_v, A, Asubdiag
        if AdiagTinv is not None:
            AdiagTinv.clean()

        return Utils.mkvc(JTv).astype(float)

    def getSourceTerm(self, tInd):
        """
        Assemble the source term. This ensures that the RHS is a vector / array
        of the correct size
        """

        Srcs = self.survey.srcList

        if self._formulation == "EB":
            s_m = np.zeros((self.mesh.nF, len(Srcs)))
            s_e = np.zeros((self.mesh.nE, len(Srcs)))
        elif self._formulation == "HJ":
            s_m = np.zeros((self.mesh.nE, len(Srcs)))
            s_e = np.zeros((self.mesh.nF, len(Srcs)))

        for i, src in enumerate(Srcs):
            smi, sei = src.eval(self, self.times[tInd])
            s_m[:, i] = s_m[:, i] + smi
            s_e[:, i] = s_e[:, i] + sei

        return s_m, s_e

    def getInitialFields(self):
        """
        Ask the sources for initial fields
        """

        Srcs = self.survey.srcList

        if self._fieldType in ["b", "j"]:
            ifields = np.zeros((self.mesh.nF, len(Srcs)))
        elif self._fieldType in ["e", "h"]:
            ifields = np.zeros((self.mesh.nE, len(Srcs)))

        for i, src in enumerate(Srcs):
            ifields[:, i] = ifields[:, i] + getattr(
                src, f"{self._fieldType}Initial", None
            )(self)

        return ifields

    def getInitialFieldsDeriv(self, src, v, adjoint=False, f=None):
        if adjoint is False:
            if self._fieldType in ["b", "j"]:
                ifieldsDeriv = np.zeros(self.mesh.nF)
            elif self._fieldType in ["e", "h"]:
                ifieldsDeriv = np.zeros(self.mesh.nE)

        elif adjoint is True:
            ifieldsDeriv = np.zeros(self.mapping.nP)

        ifieldsDeriv = (
            Utils.mkvc(
                getattr(src, f"{self._fieldType}InitialDeriv", None)(
                    self, v, adjoint, f
                )
            )
            + ifieldsDeriv
        )
        return ifieldsDeriv


###############################################################################
#                                                                             #
#                                E-B Formulation                              #
#                                                                             #
###############################################################################

# ------------------------------- Problem3D_b ------------------------------- #


class Problem3D_b(BaseTDEMProblem):
    """
    Starting from the quasi-static E-B formulation of Maxwell's equations
    (semi-discretized)

    .. math::

        \\mathbf{C} \\mathbf{e} + \\frac{\\partial \\mathbf{b}}{\\partial t} =
        \\mathbf{s_m} \\\\
        \\mathbf{C}^{\\top} \\mathbf{M_{\\mu^{-1}}^f} \\mathbf{b} -
        \\mathbf{M_{\\sigma}^e} \\mathbf{e} = \\mathbf{s_e}


    where :math:`\\mathbf{s_e}` is an integrated quantity, we eliminate
    :math:`\\mathbf{e}` using

    .. math::

        \\mathbf{e} = \\mathbf{M_{\\sigma}^e}^{-1} \\mathbf{C}^{\\top}
        \\mathbf{M_{\\mu^{-1}}^f} \\mathbf{b} -
        \\mathbf{M_{\\sigma}^e}^{-1} \\mathbf{s_e}


    to obtain a second order semi-discretized system in :math:`\\mathbf{b}`

    .. math::

        \\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1} \\mathbf{C}^{\\top}
        \\mathbf{M_{\\mu^{-1}}^f} \\mathbf{b}  +
        \\frac{\\partial \\mathbf{b}}{\\partial t} =
        \\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1} \\mathbf{s_e} + \\mathbf{s_m}


    and moving everything except the time derivative to the rhs gives

    .. math::
        \\frac{\\partial \\mathbf{b}}{\\partial t} =
        -\\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1} \\mathbf{C}^{\\top}
        \\mathbf{M_{\\mu^{-1}}^f} \\mathbf{b} +
        \\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1} \\mathbf{s_e} + \\mathbf{s_m}

    For the time discretization, we use backward euler. To solve for the
    :math:`n+1` th time step, we have

    .. math::

        \\frac{\\mathbf{b}^{n+1} - \\mathbf{b}^{n}}{\\mathbf{dt}} =
        -\\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1} \\mathbf{C}^{\\top}
        \\mathbf{M_{\\mu^{-1}}^f} \\mathbf{b}^{n+1} +
        \\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1} \\mathbf{s_e}^{n+1} +
        \\mathbf{s_m}^{n+1}


    re-arranging to put :math:`\\mathbf{b}^{n+1}` on the left hand side gives

    .. math::

        (\\mathbf{I} + \\mathbf{dt} \\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1}
         \\mathbf{C}^{\\top} \\mathbf{M_{\\mu^{-1}}^f}) \\mathbf{b}^{n+1} =
         \\mathbf{b}^{n} + \\mathbf{dt}(\\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1}
         \\mathbf{s_e}^{n+1} + \\mathbf{s_m}^{n+1})

    """

    _fieldType = "b"
    _formulation = "EB"
    fieldsPair = Fields3D_b  #: A SimPEG.EM.TDEM.Fields3D_b object
    surveyPair = SurveyTDEM

    def __init__(self, mesh, **kwargs):
        BaseTDEMProblem.__init__(self, mesh, **kwargs)

    def getAdiag(self, tInd):
        """
        System matrix at a given time index

        .. math::
            (\\mathbf{I} + \\mathbf{dt} \\mathbf{C} \\mathbf{M_{\\sigma}^e}^{-1}
            \\mathbf{C}^{\\top} \\mathbf{M_{\\mu^{-1}}^f})

        """
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]
        C = self.mesh.edgeCurl
        MeSigmaI = self.MeSigmaI
        MfMui = self.MfMui
        I = Utils.speye(self.mesh.nF)

        A = 1.0 / dt * I + (C * (MeSigmaI * (C.T * MfMui)))

        if self._makeASymmetric is True:
            return MfMui.T * A
        return A

    def getAdiagDeriv(self, tInd, u, v, adjoint=False):
        """
        Derivative of ADiag
        """
        C = self.mesh.edgeCurl

        def MeSigmaIDeriv(x):
            return self.MeSigmaIDeriv(x)

        MfMui = self.MfMui

        if adjoint:
            if self._makeASymmetric is True:
                v = MfMui * v
            return MeSigmaIDeriv(C.T * (MfMui * u)).T * (C.T * v)

        ADeriv = C * (MeSigmaIDeriv(C.T * (MfMui * u)) * v)

        if self._makeASymmetric is True:
            return MfMui.T * ADeriv
        return ADeriv

    def getAsubdiag(self, tInd):
        """
        Matrix below the diagonal
        """

        dt = self.timeSteps[tInd]
        MfMui = self.MfMui
        Asubdiag = -1.0 / dt * sp.eye(self.mesh.nF)

        if self._makeASymmetric is True:
            return MfMui.T * Asubdiag

        return Asubdiag

    def getAsubdiagDeriv(self, tInd, u, v, adjoint=False):
        return Utils.Zero() * v

    def getRHS(self, tInd):
        """
        Assemble the RHS
        """
        C = self.mesh.edgeCurl
        MeSigmaI = self.MeSigmaI
        MfMui = self.MfMui

        s_m, s_e = self.getSourceTerm(tInd)

        rhs = C * (MeSigmaI * s_e) + s_m
        if self._makeASymmetric is True:
            return MfMui.T * rhs
        return rhs

    def getRHSDeriv(self, tInd, src, v, adjoint=False):
        """
        Derivative of the RHS
        """

        C = self.mesh.edgeCurl
        MeSigmaI = self.MeSigmaI

        def MeSigmaIDeriv(u):
            return self.MeSigmaIDeriv(u)

        MfMui = self.MfMui

        _, s_e = src.eval(self, self.times[tInd])
        s_mDeriv, s_eDeriv = src.evalDeriv(self, self.times[tInd], adjoint=adjoint)

        if adjoint:
            if self._makeASymmetric is True:
                v = self.MfMui * v
            if isinstance(s_e, Utils.Zero):
                MeSigmaIDerivT_v = Utils.Zero()
            else:
                MeSigmaIDerivT_v = MeSigmaIDeriv(s_e).T * C.T * v

            RHSDeriv = MeSigmaIDerivT_v + s_eDeriv(MeSigmaI.T * (C.T * v)) + s_mDeriv(v)

            return RHSDeriv

        if isinstance(s_e, Utils.Zero):
            MeSigmaIDeriv_v = Utils.Zero()
        else:
            MeSigmaIDeriv_v = MeSigmaIDeriv(s_e) * v

        RHSDeriv = C * MeSigmaIDeriv_v + C * MeSigmaI * s_eDeriv(v) + s_mDeriv(v)

        if self._makeASymmetric is True:
            return self.MfMui.T * RHSDeriv
        return RHSDeriv


# ------------------------------- Problem3D_e ------------------------------- #
class Problem3D_e(BaseTDEMProblem):
    """
        Solve the EB-formulation of Maxwell's equations for the electric field, e.

        Starting with

        .. math::

            \\nabla \\times \\mathbf{e} + \\frac{\\partial \\mathbf{b}}{\\partial t} = \\mathbf{s_m} \\
            \\nabla \\times \\mu^{-1} \\mathbf{b} - \\sigma \\mathbf{e} = \\mathbf{s_e}


        we eliminate :math:`\\frac{\\partial b}{\\partial t}` using

        .. math::

            \\frac{\\partial \\mathbf{b}}{\\partial t} = - \\nabla \\times \\mathbf{e} + \\mathbf{s_m}


        taking the time-derivative of Ampere's law, we see

        .. math::

            \\frac{\\partial}{\\partial t}\\left( \\nabla \\times \\mu^{-1} \\mathbf{b} - \\sigma \\mathbf{e} \\right) = \\frac{\\partial \\mathbf{s_e}}{\\partial t} \\
            \\nabla \\times \\mu^{-1} \\frac{\\partial \\mathbf{b}}{\\partial t} - \\sigma \\frac{\\partial\\mathbf{e}}{\\partial t} = \\frac{\\partial \\mathbf{s_e}}{\\partial t}


        which gives us

        .. math::

            \\nabla \\times \\mu^{-1} \\nabla \\times \\mathbf{e} + \\sigma \\frac{\\partial\\mathbf{e}}{\\partial t} = \\nabla \\times \\mu^{-1} \\mathbf{s_m} + \\frac{\\partial \\mathbf{s_e}}{\\partial t}


    """

    _fieldType = "e"
    _formulation = "EB"
    fieldsPair = Fields3D_e  #: A Fields3D_e
    surveyPair = SurveyTDEM
    Adcinv = None

    def __init__(self, mesh, **kwargs):
        BaseTDEMProblem.__init__(self, mesh, **kwargs)

    def Jtvec(self, m, v, f=None):

        """
            Jvec computes the adjoint of the sensitivity times a vector

        """

        if f is None:
            f = self.fields(m)

        self.model = m
        ftype = self._fieldType + "Solution"  # the thing we solved for

        # Ensure v is a data object.
        if not isinstance(v, self.dataPair):
            v = self.dataPair(self.survey, v)

        df_duT_v = Fields_Derivs(self.mesh, self.survey)

        # same size as fields at a single timestep
        ATinv_df_duT_v = np.zeros(
            (len(self.survey.srcList), len(f[self.survey.srcList[0], ftype, 0])),
            dtype=float,
        )
        JTv = np.zeros(m.shape, dtype=float)

        # Loop over sources and receivers to create a fields object:
        # PT_v, df_duT_v, df_dmT_v
        # initialize storage for PT_v (don't need to preserve over sources)
        PT_v = Fields_Derivs(self.mesh, self.survey)
        for src in self.survey.srcList:
            # Looping over initializing field class is appending memory!
            # PT_v = Fields_Derivs(self.mesh, self.survey) # initialize storage
            # #for PT_v (don't need to preserve over sources)
            # initialize size
            df_duT_v[src, f"{self._fieldType}Deriv", :] = np.zeros_like(
                f[src, self._fieldType, :]
            )

            for rx in src.rxList:
                PT_v[src, f"{rx.projField}Deriv", :] = rx.evalDeriv(
                    src,
                    self.mesh,
                    self.timeMesh,
                    f,
                    Utils.mkvc(v[src, rx]),
                    adjoint=True,
                )
                # this is +=

                # PT_v = np.reshape(curPT_v,(len(curPT_v)/self.timeMesh.nN,
                # self.timeMesh.nN), order='F')
                df_duTFun = getattr(f, f"_{rx.projField}Deriv", None)

                for tInd in range(self.nT + 1):
                    cur = df_duTFun(
                        tInd,
                        src,
                        None,
                        Utils.mkvc(PT_v[src, f"{rx.projField}Deriv", tInd]),
                        adjoint=True,
                    )

                    df_duT_v[src, f"{self._fieldType}Deriv", tInd] = df_duT_v[
                        src, f"{self._fieldType}Deriv", tInd
                    ] + Utils.mkvc(cur[0], 2)
                    JTv = cur[1] + JTv

        # no longer need this
        del PT_v

        AdiagTinv = None

        # Do the back-solve through time
        # if the previous timestep is the same: no need to refactor the matrix
        # for tInd, dt in zip(range(self.nT), self.timeSteps):

        for tInd in reversed(range(self.nT)):
            # tInd = tIndP - 1
            if AdiagTinv is not None and (
                tInd <= self.nT and self.timeSteps[tInd] != self.timeSteps[tInd + 1]
            ):
                AdiagTinv.clean()
                AdiagTinv = None

            # refactor if we need to
            if AdiagTinv is None:  # and tInd > -1:
                Adiag = self.getAdiag(tInd)
                AdiagTinv = self.Solver(Adiag.T, **self.solverOpts)

            if tInd < self.nT - 1:
                Asubdiag = self.getAsubdiag(tInd + 1)

            for isrc, src in enumerate(self.survey.srcList):

                # solve against df_duT_v
                if tInd >= self.nT - 1:
                    # last timestep (first to be solved)
                    ATinv_df_duT_v[isrc, :] = (
                        AdiagTinv * df_duT_v[src, f"{self._fieldType}Deriv", tInd + 1]
                    )
                elif tInd > -1:
                    ATinv_df_duT_v[isrc, :] = AdiagTinv * (
                        Utils.mkvc(df_duT_v[src, f"{self._fieldType}Deriv", tInd + 1])
                        - Asubdiag.T * Utils.mkvc(ATinv_df_duT_v[isrc, :])
                    )

                dAsubdiagT_dm_v = self.getAsubdiagDeriv(
                    tInd, f[src, ftype, tInd], ATinv_df_duT_v[isrc, :], adjoint=True
                )

                dRHST_dm_v = self.getRHSDeriv(
                    tInd + 1, src, ATinv_df_duT_v[isrc, :], adjoint=True
                )  # on nodes of time mesh

                un_src = f[src, ftype, tInd + 1]
                # cell centered on time mesh
                dAT_dm_v = self.getAdiagDeriv(
                    tInd, un_src, ATinv_df_duT_v[isrc, :], adjoint=True
                )

                JTv = JTv + Utils.mkvc(-dAT_dm_v - dAsubdiagT_dm_v + dRHST_dm_v)

        # Treating initial condition when a galvanic source is included
        tInd = -1
        Grad = self.mesh.nodalGrad

        for isrc, src in enumerate(self.survey.srcList):
            if src.srcType == "Galvanic":

                ATinv_df_duT_v[isrc, :] = Grad * (
                    self.Adcinv
                    * (
                        Grad.T
                        * (
                            Utils.mkvc(
                                df_duT_v[src, f"{self._fieldType}Deriv", tInd + 1]
                            )
                            - Asubdiag.T * Utils.mkvc(ATinv_df_duT_v[isrc, :])
                        )
                    )
                )

                dRHST_dm_v = self.getRHSDeriv(
                    tInd + 1, src, ATinv_df_duT_v[isrc, :], adjoint=True
                )  # on nodes of time mesh

                un_src = f[src, ftype, tInd + 1]
                # cell centered on time mesh
                dAT_dm_v = self.MeSigmaDeriv(un_src).T * ATinv_df_duT_v[isrc, :]

                JTv = JTv + Utils.mkvc(-dAT_dm_v + dRHST_dm_v)

        # del df_duT_v, ATinv_df_duT_v, A, Asubdiag
        if AdiagTinv is not None:
            AdiagTinv.clean()

        return Utils.mkvc(JTv).astype(float)

    def getAdiag(self, tInd):
        """
        Diagonal of the system matrix at a given time index
        """
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]
        C = self.mesh.edgeCurl
        MfMui = self.MfMui
        MeSigma = self.MeSigma

        return C.T * (MfMui * C) + 1.0 / dt * MeSigma

    def getAdiagDeriv(self, tInd, u, v, adjoint=False):
        """
        Deriv of ADiag with respect to electrical conductivity
        """
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]
        MeSigmaDeriv = self.MeSigmaDeriv(u)

        if adjoint:
            return 1.0 / dt * MeSigmaDeriv.T * v

        return 1.0 / dt * MeSigmaDeriv * v

    def getAsubdiag(self, tInd):
        """
        Matrix below the diagonal
        """
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]

        return -1.0 / dt * self.MeSigma

    def getAsubdiagDeriv(self, tInd, u, v, adjoint=False):
        """
        Derivative of the matrix below the diagonal with respect to electrical
        conductivity
        """
        dt = self.timeSteps[tInd]

        if adjoint:
            return -1.0 / dt * self.MeSigmaDeriv(u).T * v

        return -1.0 / dt * self.MeSigmaDeriv(u) * v

    def getRHS(self, tInd):
        """
        right hand side
        """
        # Omit this: Note input was tInd+1
        # if tInd == len(self.timeSteps):
        #     tInd = tInd - 1

        dt = self.timeSteps[tInd - 1]
        s_m, s_e = self.getSourceTerm(tInd)
        _, s_en1 = self.getSourceTerm(tInd - 1)
        return -1.0 / dt * (s_e - s_en1) + self.mesh.edgeCurl.T * self.MfMui * s_m

    def getRHSDeriv(self, tInd, src, v, adjoint=False):
        # right now, we are assuming that s_e, s_m do not depend on the model.
        return Utils.Zero()

    def getInitialFields(self):
        """
        Ask the sources for initial fields
        """

        Srcs = self.survey.srcList

        if self._fieldType in ["b", "j"]:
            ifields = np.zeros((self.mesh.nF, len(Srcs)))
        elif self._fieldType in ["e", "h"]:
            ifields = np.zeros((self.mesh.nE, len(Srcs)))
        if self.verbose:
            print("Calculating Initial fields")

        for i, src in enumerate(Srcs):
            # Check if the source is grounded
            if src.srcType == "Galvanic" and src.waveform.hasInitialFields:
                # Check self.Adcinv and clean
                if self.Adcinv is not None:
                    self.Adcinv.clean()
                # Factorize Adc matrix
                if self.verbose:
                    print("Factorize system matrix for DC problem")
                Adc = self.getAdc()
                self.Adcinv = self.Solver(Adc)

            ifields[:, i] = ifields[:, i] + getattr(
                src, f"{self._fieldType}Initial", None
            )(self)

        return ifields

    def getAdc(self):
        MeSigma = self.MeSigma
        Grad = self.mesh.nodalGrad
        Adc = Grad.T * MeSigma * Grad
        # Handling Null space of A
        Adc[0, 0] = Adc[0, 0] + 1.0
        return Adc

    def getAdcDeriv(self, u, v, adjoint=False):
        Grad = self.mesh.nodalGrad
        if not adjoint:
            return Grad.T * (self.MeSigmaDeriv(-u) * v)
        elif adjoint:
            return self.MeSigmaDeriv(-u).T * (Grad * v)
        return Adc

    def clean(self):
        """
        Clean factors
        """
        if self.Adcinv is not None:
            self.Adcinv.clean()


###############################################################################
#                                                                             #
#                                H-J Formulation                              #
#                                                                             #
###############################################################################

# ------------------------------- Problem3D_h ------------------------------- #


class Problem3D_h(BaseTDEMProblem):
    """
    Solve the H-J formulation of Maxwell's equations for the magnetic field h.

    We start with Maxwell's equations in terms of the magnetic field and
    current density

    .. math::

        \\nabla \\times \\rho \\mathbf{j} + \\mu \\frac{\\partial h}{\\partial t} = \\mathbf{s_m} \\
        \\nabla \\times \\mathbf{h} - \\mathbf{j} = \\mathbf{s_e}


    and eliminate :math:`\\mathbf{j}` using

    .. math::

        \\mathbf{j} = \\nabla \\times \\mathbf{h} - \\mathbf{s_e}


    giving

    .. math::

        \\nabla \\times \\rho \\nabla \\times \\mathbf{h} + \\mu \\frac{\\partial h}{\\partial t} =  \\nabla \\times \\rho \\mathbf{s_e} + \\mathbf{s_m}


    """

    _fieldType = "h"
    _formulation = "HJ"
    fieldsPair = Fields3D_h  #: Fields object pair
    surveyPair = SurveyTDEM

    def __init__(self, mesh, **kwargs):
        BaseTDEMProblem.__init__(self, mesh, **kwargs)

    def getAdiag(self, tInd):
        """
        System matrix at a given time index

        """
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]
        C = self.mesh.edgeCurl
        MfRho = self.MfRho
        MeMu = self.MeMu

        return C.T * (MfRho * C) + 1.0 / dt * MeMu

    def getAdiagDeriv(self, tInd, u, v, adjoint=False):
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]
        C = self.mesh.edgeCurl
        MfRhoDeriv = self.MfRhoDeriv(C * u)

        if adjoint:
            return MfRhoDeriv.T * C * v

        return C.T * (MfRhoDeriv * v)

    def getAsubdiag(self, tInd):
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]

        return -1.0 / dt * self.MeMu

    def getAsubdiagDeriv(self, tInd, u, v, adjoint=False):
        return Utils.Zero()

    def getRHS(self, tInd):

        C = self.mesh.edgeCurl
        MfRho = self.MfRho
        s_m, s_e = self.getSourceTerm(tInd)

        return C.T * (MfRho * s_e) + s_m

    def getRHSDeriv(self, tInd, src, v, adjoint=False):
        C = self.mesh.edgeCurl
        s_m, s_e = src.eval(self, self.times[tInd])
        MfRhoDeriv = self.MfRhoDeriv(s_e)

        if adjoint is True:
            return MfRhoDeriv.T * (C * v)
        # assumes no source derivs
        return C.T * (MfRhoDeriv * v)


# ------------------------------- Problem3D_j ------------------------------- #


class Problem3D_j(BaseTDEMProblem):

    r"""
    Solve the H-J formulation for current density

    In this case, we eliminate :math:`\partial \mathbf{h} / \partial t` and
    solve for :math:`\mathbf{j}`

    """

    _fieldType = "j"
    _formulation = "HJ"
    fieldsPair = Fields3D_j  #: Fields object pair
    surveyPair = SurveyTDEM  #: survey

    def __init__(self, mesh, **kwargs):
        BaseTDEMProblem.__init__(self, mesh, **kwargs)

    def getAdiag(self, tInd):
        """
        System matrix at a given time index

        """
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]
        C = self.mesh.edgeCurl
        MfRho = self.MfRho
        MeMuI = self.MeMuI
        eye = sp.eye(self.mesh.nF)

        A = C * (MeMuI * (C.T * MfRho)) + 1.0 / dt * eye

        if self._makeASymmetric:
            return MfRho.T * A

        return A

    def getAdiagDeriv(self, tInd, u, v, adjoint=False):
        assert tInd >= 0 and tInd < self.nT

        dt = self.timeSteps[tInd]
        C = self.mesh.edgeCurl
        MfRho = self.MfRho
        MfRhoDeriv = self.MfRhoDeriv(u)
        MeMuI = self.MeMuI

        if adjoint:
            if self._makeASymmetric:
                v = MfRho * v
            return MfRhoDeriv.T * (C * (MeMuI.T * (C.T * v)))

        ADeriv = C * (MeMuI * (C.T * MfRhoDeriv * v))
        if self._makeASymmetric:
            return MfRho.T * ADeriv
        return ADeriv

    def getAsubdiag(self, tInd):
        assert tInd >= 0 and tInd < self.nT
        eye = sp.eye(self.mesh.nF)

        dt = self.timeSteps[tInd]

        if self._makeASymmetric:
            return -1.0 / dt * self.MfRho.T
        return -1.0 / dt * eye

    def getAsubdiagDeriv(self, tInd, u, v, adjoint=False):
        return Utils.Zero()

    def getRHS(self, tInd):

        if tInd == len(self.timeSteps):
            tInd = tInd - 1

        C = self.mesh.edgeCurl
        MeMuI = self.MeMuI
        dt = self.timeSteps[tInd]
        s_m, s_e = self.getSourceTerm(tInd)
        _, s_en1 = self.getSourceTerm(tInd - 1)

        rhs = -1.0 / dt * (s_e - s_en1) + C * MeMuI * s_m
        if self._makeASymmetric:
            return self.MfRho.T * rhs
        return rhs

    def getRHSDeriv(self, tInd, src, v, adjoint=False):
        return Utils.Zero()  # assumes no derivs on sources



# Analytic solution of EM fields due to a plane wave

import numpy as np, SimPEG as simpeg
from scipy.constants import mu_0, epsilon_0 as eps_0


def getEHfields(m1d, sigma, freq, zd, scaleUD=True, scaleValue=1):
    """Analytic solution for MT 1D layered earth. Returns E and H fields.

    :param SimPEG.mesh, object m1d: Mesh object with the 1D spatial information.
    :param numpy.array, vector sigma: Physical property of conductivity corresponding with the mesh.
    :param float, freq: Frequency to calculate data at.
    :param numpy array, vector zd: location to calculate EH fields at
    :param boolean, scaleUD: scales the output to be scaleValue at the top, increases numerical stability.

    Assumes a halfspace with the same conductive as the deepest cell.

    """
    # Note add an error check for the mesh and sigma are the same size.

    # Constants: Assume constant
    mu = mu_0 * np.ones(m1d.nC + 1)
    eps = eps_0 * np.ones(m1d.nC + 1)
    # Angular freq
    w = 2 * np.pi * freq
    # Add the halfspace value to the property
    sig = np.concatenate((np.array([sigma[0]]), sigma))
    # Calculate the wave number
    k = np.sqrt(eps * mu * w ** 2 - 1j * mu * sig * w)

    # Initiate the propagation matrix, in the order down up.
    UDp = np.zeros((2, m1d.nC + 1), dtype=complex)
    UDp[
        1, 0
    ] = scaleValue  # Set the wave amplitude as 1 into the half-space at the bottom of the mesh
    # Loop over all the layers, starting at the bottom layer
    for lnr, h in enumerate(m1d.hx):  # lnr-number of layer, h-thickness of the layer
        # Calculate
        yp1 = k[lnr] / (w * mu[lnr])  # Admittance of the layer below the current layer
        zp = (w * mu[lnr + 1]) / k[lnr + 1]  # Impedance in the current layer
        # Build the propagation matrix

        # Convert fields to down/up going components in layer below current layer
        Pj1 = np.array([[1, 1], [yp1, -yp1]], dtype=complex)
        # Convert fields to down/up going components in current layer
        Pjinv = 1.0 / 2 * np.array([[1, zp], [1, -zp]], dtype=complex)
        # Propagate down and up components through the current layer
        elamh = np.array(
            [[np.exp(-1j * k[lnr + 1] * h), 0], [0, np.exp(1j * k[lnr + 1] * h)]]
        )

        # The down and up component in current layer.
        UDp[:, lnr + 1] = elamh.dot(Pjinv.dot(Pj1)).dot(UDp[:, lnr])

        if scaleUD:
            # Scale the values such that 1 at the top
            scaleVal = UDp[:, lnr + 1 :: -1] / UDp[1, lnr + 1]
            if np.any(~np.isfinite(scaleVal)):
                # If there is a nan (thickness very great), rebuild the move up cell
                scaleVal = np.zeros_like(UDp[:, lnr + 1 :: -1], dtype=complex)
                scaleVal[1, 0] = scaleValue

            UDp[:, lnr + 1 :: -1] = scaleVal

    # Calculate the fields
    Ed = np.empty((zd.size,), dtype=complex)
    Eu = np.empty((zd.size,), dtype=complex)
    Hd = np.empty((zd.size,), dtype=complex)
    Hu = np.empty((zd.size,), dtype=complex)

    # Loop over the layers and calculate the fields
    # In the halfspace below the mesh
    dup = m1d.vectorNx[0]
    dind = dup >= zd
    Ed[dind] = UDp[1, 0] * np.exp(-1j * k[0] * (dup - zd[dind]))
    Eu[dind] = UDp[0, 0] * np.exp(1j * k[0] * (dup - zd[dind]))
    Hd[dind] = (k[0] / (w * mu[0])) * UDp[1, 0] * np.exp(-1j * k[0] * (dup - zd[dind]))
    Hu[dind] = -(k[0] / (w * mu[0])) * UDp[0, 0] * np.exp(1j * k[0] * (dup - zd[dind]))
    for ki, mui, epsi, dlow, dup, Up, Dp in zip(
        k[1::],
        mu[1::],
        eps[1::],
        m1d.vectorNx[:-1],
        m1d.vectorNx[1::],
        UDp[0, 1::],
        UDp[1, 1::],
    ):
        dind = np.logical_and(dup >= zd, zd > dlow)
        Ed[dind] = Dp * np.exp(-1j * ki * (dup - zd[dind]))
        Eu[dind] = Up * np.exp(1j * ki * (dup - zd[dind]))
        Hd[dind] = (ki / (w * mui)) * Dp * np.exp(-1j * ki * (dup - zd[dind]))
        Hu[dind] = -(ki / (w * mui)) * Up * np.exp(1j * ki * (dup - zd[dind]))

    # Return return the fields
    return Ed, Eu, Hd, Hu


def getImpedance(m1d, sigma, freq):
    """Analytic solution for MT 1D layered earth. Returns the impedance at the surface.

    :param SimPEG.mesh, object m1d: Mesh object with the 1D spatial information.
    :param numpy.array, vector sigma: Physical property corresponding with the mesh.
    :param numpy.array, vector freq: Frequencies to calculate data at.


    """

    # Initiate the impedances
    Z1d = np.empty(len(freq), dtype="complex")
    h = m1d.hx  # vectorNx[:-1]
    # Start the process
    for nrFr, fr in enumerate(freq):
        om = 2 * np.pi * fr
        Zall = np.empty(len(h) + 1, dtype="complex")
        # Calculate the impedance for the bottom layer
        Zall[0] = (mu_0 * om) / np.sqrt(
            mu_0 * eps_0 * (om) ** 2 - 1j * mu_0 * sigma[0] * om
        )

        for nr, hi in enumerate(h):
            # Calculate the wave number
            # print(nr,sigma[nr])
            k = np.sqrt(mu_0 * eps_0 * om ** 2 - 1j * mu_0 * sigma[nr] * om)
            Z = (mu_0 * om) / k

            Zall[nr + 1] = Z * (
                (Zall[nr] + Z * np.tanh(1j * k * hi))
                / (Z + Zall[nr] * np.tanh(1j * k * hi))
            )

        # pdb.set_trace()
        Z1d[nrFr] = Zall[-1]

    return Z1d



from discretize.base import BaseMesh
from . import Utils
from . import Survey
from . import Models
import numpy as np
from . import Maps
from .Fields import Fields, TimeFields
from . import Mesh
from . import Props
import properties


Solver = Utils.SolverUtils.Solver


class BaseProblem(Props.HasModel):
    """Problem is the base class for all geophysical forward problems
    in SimPEG.
    """

    #: A SimPEG.Utils.Counter object
    counter = None

    #: A SimPEG.Survey Class
    surveyPair = Survey.BaseSurvey

    #: A SimPEG.Map Class
    mapPair = Maps.IdentityMap

    #: A SimPEG Solver class.
    Solver = Solver

    #: Solver options as a kwarg dict
    solverOpts = {}

    #: A discretize instance.
    mesh = None

    def __init__(self, mesh, **kwargs):

        # raise exception if user tries to set "mapping"
        if "mapping" in kwargs:
            raise Exception(
                "Depreciated (in 0.4.0): use one of {}".format(
                    [p for p in self._props.keys() if "Map" in p]
                )
            )

        super().__init__(**kwargs)
        assert isinstance(mesh, BaseMesh), "mesh must be a discretize object."
        self.mesh = mesh

    @property
    def mapping(self):
        """Setting an unnamed mapping has been depreciated in
        v0.4.0. Please see the release notes for more details.
        """
        raise Exception(
            "Depreciated (in 0.4.0): use one of {}".format(
                [p for p in self._props.keys() if "Map" in p]
            )
        )

    @mapping.setter
    def mapping(self, value):
        raise Exception(
            "Depreciated (in 0.4.0): use one of {}".format(
                [p for p in self._props.keys() if "Map" in p]
            )
        )

    @property
    def curModel(self):
        """
        Setting the curModel is depreciated.

        Use `SimPEG.Problem.model` instead.
        """
        raise AttributeError(
            "curModel is depreciated (in 0.4.0). Use " "`SimPEG.Problem.model` instead"
        )

    @curModel.setter
    def curModel(self, value):
        raise AttributeError(
            "curModel is depreciated (in 0.4.0). Use " "`SimPEG.Problem.model` instead"
        )

    @property
    def survey(self):
        """
        The survey object for this problem.
        """
        return getattr(self, "_survey", None)

    def pair(self, d):
        """Bind a survey to this problem instance using pointers."""
        assert isinstance(
            d, self.surveyPair
        ), "Data object must be an instance of a {!s} class.".format(
            self.surveyPair.__name__
        )
        if d.ispaired:
            raise Exception(
                "The survey object is already paired to a problem. "
                "Use survey.unpair()"
            )
        self._survey = d
        d._prob = self

    def unpair(self):
        """Unbind a survey from this problem instance."""
        if not self.ispaired:
            return
        self.survey._prob = None
        self._survey = None

    #: List of strings, e.g. ['_MeSigma', '_MeSigmaI']
    deleteTheseOnModelUpdate = []

    #: List of matrix names to have their factors cleared on a model update
    clean_on_model_update = []

    @properties.observer("model")
    def _on_model_update(self, change):
        if change["previous"] is change["value"]:
            return
        if (
            isinstance(change["previous"], np.ndarray)
            and isinstance(change["value"], np.ndarray)
            and np.allclose(change["previous"], change["value"])
        ):
            return

        for prop in self.deleteTheseOnModelUpdate:
            if hasattr(self, prop):
                delattr(self, prop)

        # matrix factors to clear
        for mat in self.clean_on_model_update:
            if getattr(self, mat, None) is not None:
                getattr(self, mat).clean()  # clean factors
                setattr(self, mat, None)  # set to none

    @property
    def ispaired(self):
        """True if the problem is paired to a survey."""
        return self.survey is not None

    @Utils.timeIt
    def Jvec(self, m, v, f=None):
        """Jvec(m, v, f=None)

        Effect of J(m) on a vector v.

        :param numpy.array m: model
        :param numpy.array v: vector to multiply
        :param Fields f: fields
        :rtype: numpy.array
        :return: Jv
        """
        raise NotImplementedError("J is not yet implemented.")

    @Utils.timeIt
    def Jtvec(self, m, v, f=None):
        """Jtvec(m, v, f=None)

        Effect of transpose of J(m) on a vector v.

        :param numpy.array m: model
        :param numpy.array v: vector to multiply
        :param Fields f: fields
        :rtype: numpy.array
        :return: JTv
        """
        raise NotImplementedError("Jt is not yet implemented.")

    @Utils.timeIt
    def Jvec_approx(self, m, v, f=None):
        """Jvec_approx(m, v, f=None)

        Approximate effect of J(m) on a vector v

        :param numpy.array m: model
        :param numpy.array v: vector to multiply
        :param Fields f: fields
        :rtype: numpy.array
        :return: approxJv
        """
        return self.Jvec(m, v, f)

    @Utils.timeIt
    def Jtvec_approx(self, m, v, f=None):
        """Jtvec_approx(m, v, f=None)

        Approximate effect of transpose of J(m) on a vector v.

        :param numpy.array m: model
        :param numpy.array v: vector to multiply
        :param Fields f: fields
        :rtype: numpy.array
        :return: JTv
        """
        return self.Jtvec(m, v, f)

    def fields(self, m):
        """The field given the model.

        :param numpy.array m: model
        :rtype: numpy.array
        :return: u, the fields
        """
        raise NotImplementedError("fields is not yet implemented.")


class BaseTimeProblem(BaseProblem):
    """Sets up that basic needs of a time domain problem."""

    @property
    def timeSteps(self):
        """Sets/gets the timeSteps for the time domain problem.

        You can set as an array of dt's or as a list of tuples/floats.
        Tuples must be length two with [..., (dt, repeat), ...]

        For example, the following setters are the same::

            prob.timeSteps = [(1e-6, 3), 1e-5, (1e-4, 2)]
            prob.timeSteps = np.r_[1e-6,1e-6,1e-6,1e-5,1e-4,1e-4]

        """
        return getattr(self, "_timeSteps", None)

    @timeSteps.setter
    def timeSteps(self, value):
        if isinstance(value, np.ndarray):
            self._timeSteps = value
            del self.timeMesh
            return

        self._timeSteps = Utils.meshTensor(value)
        del self.timeMesh

    @property
    def nT(self):
        "Number of time steps."
        return self.timeMesh.nC

    @property
    def t0(self):
        return getattr(self, "_t0", 0.0)

    @t0.setter
    def t0(self, value):
        assert np.isscalar(value), "t0 must be a scalar"
        del self.timeMesh
        self._t0 = float(value)

    @property
    def times(self):
        "Modeling times"
        return self.timeMesh.vectorNx

    @property
    def timeMesh(self):
        if getattr(self, "_timeMesh", None) is None:
            self._timeMesh = Mesh.TensorMesh([self.timeSteps], x0=[self.t0])
        return self._timeMesh

    @timeMesh.deleter
    def timeMesh(self):
        if hasattr(self, "_timeMesh"):
            del self._timeMesh


class LinearProblem(BaseProblem):

    # model, modelMap, modelDeriv = Props.Invertible(
    #     "Generic model parameters",
    #     default=1.
    # )

    G = None

    def __init__(self, mesh, **kwargs):
        BaseProblem.__init__(self, mesh, **kwargs)
        self.modelMap = kwargs.pop("mapping", Maps.IdentityMap(mesh))

    @property
    def modelMap(self):
        "A SimPEG.Map instance."
        return getattr(self, "_modelMap", None)

    @modelMap.setter
    def modelMap(self, val):
        val._assertMatchesPair(self.mapPair)
        self._modelMap = val

    def fields(self, m):
        return self.G.dot(self.modelMap * m)

    def getJ(self, m, f=None):
        """
            Sensitivity matrix
        """

        if self.modelMap is not None:
            dmudm = self.modelMap.deriv(m)
            return self.G * dmudm
        else:
            return self.G

    def Jvec(self, m, v, f=None):
        return self.G.dot(self.modelMap.deriv(m) * v)

    def Jtvec(self, m, v, f=None):
        return self.modelMap.deriv(m).T * self.G.T.dot(v)
